/* abstract-document.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <gege2061@redaction-developpez.com>
 */

public abstract class Valide.AbstractTextDocument : Gtk.VBox, iDocument
{
  public SplitSourceView split_view;
  private bool _save;
  private Searching searching;
  private uint64 mtime;
  private DocumentState _state;
  private GotoDialog goto_dialog = null;

  public signal void cursor_moved (Gtk.TextBuffer buffer);

  protected abstract SplitSourceView create_source_view ();

  public Gtk.SourceBuffer buffer
  {
    get
    {
      return (Gtk.SourceBuffer)this.text_view.buffer;
    }
  }

  public Gtk.SourceView text_view
  {
    get
    {
      return this.split_view.active_view;
    }
  }

  public Gtk.SourceView view1
  {
    get
    {
      return this.split_view.view1;
    }
  }

  public Gtk.SourceView view2
  {
    get
    {
      return this.split_view.view2;
    }
  }

  public bool is_save
  {
    get
    {
      return this._save;
    }
    set
    {
      this._save = value;
      this.title_change ();
    }
  }

  public new DocumentState state
  {
    get
    {
      return this._state;
    }
    set
    {
      this._state = value;
      this.title_change ();
      switch (this.state)
      {
        case DocumentState.NORMAL:
    	    this.editable = true;
        break;
        case DocumentState.EXTERNALLY_MODIFIED:
    	    this.editable = false;
        break;
      }
      this.state_change (this.state);
    }
  }

  public bool editable
  {
    get
    {
      return this.text_view.get_editable ();
    }
    set
    {
      this.text_view.set_editable (value);
    }
  }

  public string selected_text
  {
    owned get
    {
      string text = "";
      Gtk.TextIter start;
      Gtk.TextIter end;

      if (this.buffer.get_selection_bounds (out start, out end))
      {
        text = this.buffer.get_text (start, end, false);
      }
      return text;
    }
  }

  public new string path
  {
    owned get;
    set;
    default = _("New file");
  }

  public string filename
  {
    owned get
    {
      return Path.get_basename (this.path);
    }
  }

  public string title
  {
    owned get
    {
      string text = "";

      if (!this.is_save)
      {
        text = "* ";
      }
      text += this.filename;
      return text;
    }
  }

  private void check_externally_modified ()
  {
    if (this.path != _("New file") && this.state == DocumentState.NORMAL)
    {
      if (this.mtime < Utils.get_mtime (this.path))
      {
        this.state = DocumentState.EXTERNALLY_MODIFIED;

        DocumentMessage msg = new DocumentMessage (this);
        this.pack_start (msg, false, true, 0);
        this.reorder_child (msg, 0);
        msg.show_all ();
      }
    }
  }

  [Callback]
  private void update_cursor_position_statusbar ()
  {
	  int row;
	  int col;
    Gtk.TextIter iter;
	  Gtk.TextIter start;
	  uint tab_size;

    buffer.get_iter_at_mark (out iter, this.buffer.get_insert ());

    row = iter.get_line ();

    start = iter;
    start.set_line_offset (0);
    col = 0;

    tab_size = this.text_view.get_tab_width ();

    while (!start.equal (iter))
    {
      /* FIXME: Are we Unicode compliant here? */
      if (start.get_char () == '\t')
      {
        col += (int)(tab_size - (col  % tab_size));
      }
      else
      {
        col++;
      }

      start.forward_char ();
    }

    Statusbar.get_instance ().set_cursor_position (row + 1, col + 1);
  }

  protected virtual string get_contents () throws Error
  {
    string contents;

    FileUtils.get_contents (this.path, out contents);
    return contents;
  }

  protected virtual void set_contents (string contents) throws Error
  {
    FileUtils.set_contents (this.path, contents, -1);
  }

  construct
  {
    this.split_view = this.create_source_view ();
    this.pack_start (this.split_view, true, true, 0);
    this.cursor_moved += this.update_cursor_position_statusbar;
    this.buffer.changed += () => {
      this.is_save = false;
      this.cursor_moved (this.buffer);
    };
    this.buffer.mark_set += () => {
      this.cursor_moved (this.buffer);
    };

    this.searching = new Searching (this);
    this.pack_start (this.searching, false, true, 0);
    this.searching.hide ();

    this.update_cursor_position_statusbar ();

    this.state = DocumentState.NORMAL;
    this.text_view.focus_in_event += () => {
      this.check_externally_modified ();
      this.update_cursor_position_statusbar ();
      return false;
    };
  }

  public new void grab_focus ()
  {
    this.text_view.grab_focus ();
  }

  public new void show ()
  {
    this.show_all ();
    this.searching.hide ();
  }

  public override void show_all ()
  {
    base.show_all ();
    this.searching.hide ();
  }

  public void reload () throws Error
  {
    string contents;
    Gtk.TextIter start;
    Gtk.TextIter end;

    contents = this.get_contents ();
    this.buffer.get_start_iter (out start);
    this.buffer.get_end_iter (out end);
    this.buffer.begin_not_undoable_action ();
    this.buffer.delete (start, end);
    this.buffer.insert (start, contents, -1);
    this.buffer.end_not_undoable_action ();
    this.is_save = true;
    this.mtime = Utils.get_mtime (this.path);
  }

  public void open (string uri) throws Error
  {
    this.path = uri;
    this.reload ();
  }

  public void save ()
  {
    if (!this.is_save)
    {
      if (this.path == null)
      {
        var dialog = new Gtk.FileChooserDialog (_("Save file"), null,
                                                Gtk.FileChooserAction.SAVE,
                                                Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                                Gtk.STOCK_SAVE, Gtk.ResponseType.ACCEPT,
                                                null);
        if (dialog.run () == Gtk.ResponseType.ACCEPT)
        {
          this.path = dialog.get_filename ();
        }
        dialog.destroy ();
      }
      if (this.path != null)
      {
        Gtk.TextIter start;
        Gtk.TextIter end;
        string contents;

        this.buffer.get_bounds (out start, out end);
        contents = this.buffer.get_text (start, end, false);
        try
        {
          this.set_contents (contents);
          this.mtime = Utils.get_mtime (this.path);
          this.is_save = true;
          this.saved ();
        }
        catch (Error e)
        {
          warning (e.message);
        }
      }
    }
  }

  public void save_as ()
  {
    bool old_is_save = this.is_save;
    string old_path = this.path;

    this.path = null;
    this.is_save = false;
    this.save ();
    if (!this.is_save)
    {
      this.is_save = old_is_save;
      this.path = old_path;
    }
  }

  public bool close ()
  {
    bool is_close = true;

    if (!this.is_save)
    {
      Gtk.Dialog dialog = new Gtk.Dialog ();
      dialog.set_title (_("Are you sure?"));
      dialog.add_button (Gtk.STOCK_YES, Gtk.ResponseType.YES);
      dialog.add_button (Gtk.STOCK_NO, Gtk.ResponseType.NO);
      dialog.add_button (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL);
      dialog.set_modal (true);

      Gtk.Label label = new Gtk.Label (_("Document with unsaved changes. Save changes before closing?"));
      weak Gtk.Box vbox = (Gtk.Box)dialog.vbox;
      vbox.pack_start (label, true, false, 10);
      dialog.show_all ();
      switch (dialog.run ())
      {
        case Gtk.ResponseType.YES:
          this.save ();
          is_close = this.is_save;
        break;
        case Gtk.ResponseType.NO:
        break;
        case Gtk.ResponseType.CANCEL:
          is_close =  false;
        break;
      }
      dialog.destroy ();
    }
    if (is_close)
    {
      this.closed ();
    }
    return is_close;
  }

  public void redo ()
  {
    this.buffer.redo ();
  }

  public void search ()
  {
    this.searching.find ();
  }

  public void find_next ()
  {
    this.searching.find_next ();
  }

  public void find_prev ()
  {
    this.searching.find_prev ();
  }

  public bool goto_line (int line = -1)
  {
    bool ret = false;
	  uint line_count;
	  Gtk.TextIter iter;

	  if (line < 0)
	  {
	    if (this.goto_dialog == null)
	    {
        this.goto_dialog = new GotoDialog (this);

        Gtk.Window toplevel = (Gtk.Window)this.get_ancestor (typeof (Gtk.Window));
        if (toplevel != null)
        {
          Gtk.WindowGroup group = toplevel.group;
          if (group != null)
          {
            group.add_window (this.goto_dialog);
          }
        }
      }
	    this.goto_dialog.show ();
	  }
    else
	  {
    	line_count = this.buffer.get_line_count ();

    	if (line <= line_count)
	    {
	    	this.buffer.get_iter_at_line (out iter, line);
      	this.buffer.place_cursor (iter);

        Gtk.TextMark mark = this.buffer.create_mark ("cursor", iter, true);
        this.text_view.scroll_to_mark (mark, 0.4, false, 0, 0);
        this.buffer.delete_mark (mark);
        
	    	ret = true;
	    }
    }
    return ret;
  }

  public virtual Gdk.Pixbuf get_icon ()
  {
    Gdk.Pixbuf pixbuf;

    switch (this.state)
    {
      case DocumentState.NORMAL:
        pixbuf = Utils.get_pixbuf_for_stock (Gtk.STOCK_FILE, Gtk.IconSize.MENU);
      break;
      case DocumentState.EXTERNALLY_MODIFIED:
        pixbuf = Utils.get_pixbuf_for_stock (Gtk.STOCK_DIALOG_WARNING, Gtk.IconSize.MENU);
      break;
      default:
        pixbuf = Utils.get_pixbuf_for_stock (Gtk.STOCK_DIALOG_QUESTION, Gtk.IconSize.MENU);
      break;
    }
    return pixbuf;
  }

  public void select_range (Position pos)
  {
    Gtk.TextIter start;
    Gtk.TextIter end;

    if (pos.start_line >= 0 && pos.end_line >= 0
        && pos.start_row >= 0 && pos.end_row >= 0)
    {
      this.buffer.get_start_iter (out start);
      end = start;
      start.set_line (pos.start_line);
      start.set_line_offset (pos.start_row);
      end.set_line (pos.end_line);
      end.set_line_offset (pos.end_row);

      this.buffer.select_range (start, end);
    }
    else
    {
      debug (_("Invalide range"));
    }
  }
}

