/* compiler.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <gege2061@redaction-developpez.com>
 */

errordomain Valide.ErrorParsing
{
	FILE,
	LINE
}

public class Valide.Compiler : AbstractExecutable
{
  /**
   * Directory where executable was launched.
   */
  protected override string working_directory
  {
    owned get { return ""; }
  }

  /**
   * Message show when executable start.
   */
  protected override string start_msg
  {
    owned get { return _("<b>Compiling start</b>"); }
  }

  /**
   * Message show when executable stop.
   */
  protected override string end_msg
  {
    owned get { return _("<b>Compiling end</b>"); }
  }

  /**
   * Executable command.
   */
  protected override string executable
  {
    owned get { return "valac -q"; }
    set {}
  }

  /**
   * Return files in selected line for go to error/warning file.
   *
   * @param line Selected line in ExecutableManager.
   *
   * @return Corresponding file.
   */
  private string get_file (string line) throws Error
  {
    string[] fields = line.split (":", -1);
    if (strv_length (fields) < 2)
    {
    	throw new ErrorParsing.FILE (line);
    }
    return fields[0];
  }

  /**
   * Return line number for the selected line (used to implement "go to error/warning" etc).
   *
   * @param line Selected line in ExecutableManager.
   *
   * @return Corresponding line.
   */
  private Position get_line (string line) throws Error
  {
    Position pos = Position ();
    MatchInfo match_info;

    Regex regex = new Regex ("([\\d]+)", //"([\\d]+)\\.([\\d]+)-([\\d]+)\\.([\\d]+)",
                             RegexCompileFlags.OPTIMIZE,
                             RegexMatchFlags.NOTEMPTY);

    regex.match (line, 0, out match_info);

    int i = 0;
    while (match_info.matches ())
    {
      switch (i)
      {
        case 0:
          pos.start_line = match_info.fetch (0).to_int () - 1;
        break;
        case 1:
          pos.start_row = match_info.fetch (0).to_int () - 1;
        break;
        case 2:
          pos.end_line = match_info.fetch (0).to_int () - 1;
        break;
        case 3:
          pos.end_row = match_info.fetch (0).to_int ();
        break;
      }
      i++;
      match_info.next ();
    }
    return pos;
  }

  /**
   * Called when new line is added in ExecutableManager for parsing.
   *
   * @param sender The object which emit the signal.
   * @param line New line.
   */
  protected override void parsing (ExecutableManager sender, ref string line)
  {
     if (line.has_prefix ("Compilation failed"))
     {
       line = "<span foreground=\"#FF0000\">" + line + "</span>";
     }
     else if (line.str ("error:") != null)
     {
       line = "<span foreground=\"#FF0000\">" + line + "</span>";
     }
     if (line.has_prefix ("Compilation succeeded"))
     {
       line = "<span foreground=\"#ED7F10\">" + line + "</span>";
     }
     else if (line.str ("warning:") != null)
     {
       line = "<span foreground=\"#ED7F10\">" + line + "</span>";
     }
  }

  /**
   * Called when line clicked in ExecutableManager.
   *
   * @param sender The object which emit the signal.
   * @param line Selected line.
   */
  protected override void line_activated (ExecutableManager sender, string line)
  {
    try
    {
    	string file = this.get_file (line);
      ProjectManager.get_instance ().current.open_file (file);

      try
      {
			  Position pos = this.get_line (line);
        iDocument document = DocumentManager.get_instance ().current;
        /* Only text documents have lines */
        if (document is AbstractTextDocument)
        {
          ((AbstractTextDocument)document).select_range (pos);
          ((AbstractTextDocument)document).grab_focus ();
        }
      }
      catch (Error e)
      {
        debug (e.message);
      }
    }
    catch (Error e)
    {
      warning (e.message);
    }
  }
}

