/* document-manager.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <gege2061@redaction-developpez.com>
 */

public errordomain Valide.DocumentError
{
  UNKNOW_DOCUMENT,
  BAD_URI,
  NO_DOCUMENT
}

private class Valide.CloseDialog : AbstractCloseDialog
{
  private Gtk.ListStore list_store;

  /* enum for tree model columns */
  enum Col
  {
	  ACTIVE,
	  TITLE,
	  DOCUMENT,
    NB
  }

  public List<iDocument> selected
  {
    owned get
    {
      Gtk.TreeIter iter;
      List<iDocument> documents = new List<iDocument> ();

      this.list_store.get_iter_first (out iter);
      do
      {
        bool active;
        iDocument document;

        this.list_store.get (iter, Col.ACTIVE, out active, Col.DOCUMENT, out document);
        if (active)
        {
          documents.append (document);
        }
      } while (this.list_store.iter_next (ref iter));
      return documents;
    }
  }

  [Callback]
  private void active_toggled (Gtk.CellRendererToggle sender, string path_str)
  {
    bool active;
    Gtk.TreeIter iter;
    Gtk.TreePath path;

    path = new Gtk.TreePath.from_string (path_str);
    this.list_store.get_iter (out iter, path);
    this.list_store.get (iter, Col.ACTIVE, out active);

    active = !active;

    this.list_store.set (iter, Col.ACTIVE, active);
  }

  public CloseDialog (List<iDocument> documents)
  {
    int unsave = 0;

    foreach (iDocument document in documents)
    {
      if (!document.is_save)
      {
        unsave++;
      }
    }

    this.widgets.label.set_markup (_("<b>There are %d documents with unsaved changes. Save changes before closing?</b>").printf (unsave));

    this.list_store = new Gtk.ListStore (Col.NB, typeof (bool),
                                                 typeof (string),
                                                 typeof (iDocument));
    this.widgets.tree_view.model = this.list_store;

    Gtk.CellRendererToggle toggle_render = new Gtk.CellRendererToggle ();
    toggle_render.toggled += active_toggled;
    Gtk.TreeViewColumn column = new Gtk.TreeViewColumn.with_attributes (_("Save"), toggle_render, "active", Col.ACTIVE, null);
    this.widgets.tree_view.append_column (column);

    Gtk.CellRendererText text_render = new Gtk.CellRendererText ();
    column = new Gtk.TreeViewColumn.with_attributes (_("Filename"), text_render, "text", Col.TITLE, null);
    this.widgets.tree_view.append_column (column);

    foreach (iDocument document in documents)
    {
      if (!document.is_save)
      {
        Gtk.TreeIter iter;
        string path = document.title;

        if (path.has_prefix ("* "))
        {
          path = path.substring (2, path.len () - 2);
        }
        this.list_store.append (out iter);
        this.list_store.set (iter, Col.ACTIVE, true, Col.TITLE, path, Col.DOCUMENT, document);
      }
    }
  }

  public uint run ()
  {
    this.widgets.dialog.show_all ();
    return this.widgets.dialog.run ();
  }

  public new void destroy ()
  {
    this.widgets.dialog.destroy ();
  }
}

public class Valide.DocumentManager : Gtk.Notebook
{
  private static DocumentManager instance = null;
  private StartPage start_page;

  /** New document is adding */
  public signal void tab_added (iDocument document);
  /** Current document is saving */
  public signal void tab_saved ();
  /** Current document is closing */
  public signal void tab_removed ();
  /** Current document changed */
  public signal void tab_state_changed (iDocument document);

  /**
   * The current opened document.
   */
  public iDocument current
  {
    get
    {
       return (iDocument)this.get_nth_page (this.page);
    }
    set
    {
      this.set_current_page (this.page_num (value));
    }
  }

  /**
   * All opened documents.
   */
  public weak List<iDocument> documents
  {
    get;
    private set;
  }

  [Callback]
  private void title_change (iDocument sender)
  {
    Gtk.HBox title = new Gtk.HBox (false, 4);
    Gtk.Image image = new Gtk.Image.from_pixbuf (sender.get_icon ());
    title.pack_start (image, false, true, 0);

    Gtk.Label label = new Gtk.Label (sender.title);
    title.pack_start (label, false, false, 0);

    /* Close button */
  	image = new Gtk.Image.from_stock (Gtk.STOCK_CLOSE, Gtk.IconSize.MENU);
	  Gtk.Button button = Utils.create_small_button (image);
    button.set_data ("document", sender);
    button.clicked += (s) => {
      iDocument document = (iDocument)s.get_data ("document");
      document.close ();
    };
    button.set_tooltip_text (_("Close document"));
    title.pack_start (button, false, false, 0);

    title.show_all ();
    this.set_tab_label (sender, title);
  }

  protected virtual void setup_ui ()
  {
    bool active = false;
    UIManager ui = UIManager.get_instance ();

    if (this.current != null)
    {
      active = true;
    }
    ui.get_action ("file-save").sensitive = active;
    ui.get_action ("file-save-as").sensitive = active;
    ui.get_action ("file-close").sensitive = active;
    ui.get_action ("file-close-all").sensitive = active;
    ui.get_action ("edit-undo").sensitive = active;
    ui.get_action ("edit-redo").sensitive = active;
    ui.get_action ("edit-copy").sensitive = active;
    ui.get_action ("edit-cut").sensitive = active;
    ui.get_action ("edit-paste").sensitive = active;
    ui.get_action ("edit-search").sensitive = active;
    ui.get_action ("edit-find-next").sensitive = active;
    ui.get_action ("edit-find-prev").sensitive = active;
    ui.get_action ("edit-goto-line").sensitive = active;
  }

  /* @todo Use plugins */
  protected virtual AbstractTextDocument create_document (string? filename) throws DocumentError
  {
    string ext = "Unknow";

    if (filename != null)
    {
      string tmp = Utils.get_extension (filename);
      if (tmp != null)
      {
        ext = tmp.up (1) + tmp.offset (1);
      }
    }
    else
    {
      ext = "Vala";
    }
    Type type = 0;

    /* Search standard behavior */
    type = Type.from_name ("Valide" + ext + "Document");
    /* Or, open as text */
    if (type == 0)
    {
      string mime = Utils.get_mime_type (filename);
      if (mime.has_prefix ("text/"))
      {
        type = typeof (TxtDocument);
      }

      /* Default system behavior */
      if (type == 0)
      {
        /* @todo Use external program */
        throw new DocumentError.UNKNOW_DOCUMENT (_("Couldn't open this document"));
      }
    }
    return Object.new (type) as AbstractTextDocument;
  }

  private DocumentManager ()
  {
  }

  construct
  {
    this.start_page = new StartPage ();
    this.start_page.closed += () => {
      this.hide_start_page ();
      ((Gtk.ToggleAction)UIManager.get_instance ()
          .get_action ("view-start-page")).set_active (false);
    };
    ((Gtk.ToggleAction)UIManager.get_instance ()
        .get_action ("view-start-page")).set_active (true);
    this.show_start_page ();
    this.setup_ui ();
  }

  /**
   * Get uniq instance of DocumentManager.
   *
   * @return The DocumentManager.
   */
  public static DocumentManager get_instance ()
  {
    if (instance == null)
    {
      instance = new DocumentManager ();

      /* Force file include */
      ValaDocument tmp = null;
      /* Force type register */
      typeof (TxtDocument);
      typeof (ValaDocument);
      typeof (VapiDocument);
      typeof (XmlDocument);
      typeof (GladeDocument);
      typeof (UiDocument);
    }
    return instance;
  }

  /* @todo Add document type parameter. */
  /**
   * Create a new document.
   *
   * @param filename: a name of a file or null for create a new document.
   *
   * @return The new document.
   *
   */
  public iDocument create (string? filename = null) throws DocumentError
  {
    int pos;

    if (!this.is_open (filename, out pos))
    {
      AbstractTextDocument document;

      document = this.create_document (filename);
      if (filename != null)
      {
        document.open (filename);
        RecentManagerUI.get_instance ().add_file (filename);
      }
      document.title_change += this.title_change;
      this.append_page (document);
      document.show ();
      this.current = document;
      this.current.grab_focus ();
      this.current.closed += (s) => {
        this.remove_page (s);
      };
      this.current.saved += (s) => {
        this.tab_saved ();
      };
      this.current.state_change += (s) => {
        this.tab_state_changed (s);
      };
    }
    else
    {
      this.set_current_page (pos);
    }
    return this.current;
  }

  /**
   * Open document with #Gtk.FileChosserDialog.
   */
  public void open () throws DocumentError
  {
    Gtk.FileChooserDialog dialog;

    dialog = new Gtk.FileChooserDialog (_("Open file"), null,
                                        Gtk.FileChooserAction.OPEN,
                                        Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                        Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT,
                                        null);
    if (dialog.run () == Gtk.ResponseType.ACCEPT)
    {
      this.create (dialog.get_filename ());
      this.current.is_save = true;
    }
    dialog.destroy ();
  }

  /**
   * Close all opeining documents.
   *
   * @return true if all documents are really closed.
   */
  public bool close_all ()
  {
    int unsave = 0;
    bool close = true;

    foreach (iDocument document in documents)
    {
      if (!document.is_save)
      {
        unsave++;
      }
    }

    if (unsave > 0)
    {
      CloseDialog dialog = new CloseDialog (this.documents);

      switch (dialog.run ())
      {
        case Gtk.ResponseType.OK:
          foreach (iDocument document in dialog.selected)
          {
            document.save ();
          }
          close = true;
        break;
        case Gtk.ResponseType.CLOSE:
          close = true;
        break;
        case Gtk.ResponseType.CANCEL:
          close = false;
        break;
      }
      dialog.destroy ();
    }

    if (close)
    {
      List<iDocument> documents = this.documents.copy ();
      foreach (iDocument document in documents)
      {
        document.is_save = true;
        document.close ();
      }
    }
    return (this.get_n_pages () == 0);
  }

  /**
   * Append new page in notebook.
   *
   * @param child: the document added.
   */
  public new int append_page (iDocument child)
  {
    int page_num;

    this.documents.append (child);
    page_num = base.append_page (child, null);
    this.title_change (child);
    this.setup_ui ();
    this.tab_added (child);
    return page_num;
  }

  /**
   * Remove page from notebook.
   *
   * @param child: a document.
   */
  public new void remove_page (iDocument child)
  {
    int page_num = this.page_num (child);

    if (page_num >= 0)
    {
      base.remove_page (page_num);
      this.documents.remove (child);
      this.setup_ui ();
      this.tab_removed ();
    }
  }

  /**
   * Show the start page.
   */
  public void show_start_page ()
  {
    int page_num = this.page_num (this.start_page);

    if (page_num < 0)
    {
      this.append_page (this.start_page);
      this.start_page.show_all ();
    }
    this.current = this.start_page;
  }

  /**
   * Hide the start page.
   */
  public void hide_start_page ()
  {
    this.remove_page (this.start_page);
  }

  /**
   * Check if file is already open.
   *
   * @param filename: the path of file.
   * @param pos: return tab position.
   *
   * @return true if file is already open.
   */
  public bool is_open (string? filename, out int pos)
  {
    bool opened = false;

    if (filename != null)
    {
      pos = 0;
      foreach (iDocument d in this.documents)
      {
        if (d.path == filename)
        {
          opened = true;
          break;
        }
        pos++;
      }
    }
    return opened;
  }
}

