/* project-dialog.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <gege2061@redaction-developpez.com>
 */

private enum Valide.ColumnType
{
  ICON,
  NAME,
  DESCRIPTION,
  PATH,
  NB_COLUMNS
}

public class Valide.ProjectDialog : AbstractProjectDialog
{
  private Gtk.ListStore list_store;
  private Gtk.ListStore licenses_list;

  public signal void apply ();

  public string project_name
  {
    owned get
    {
      return this.widgets.name_entry.text;
    }
  }

  public string project_author
  {
    owned get
    {
      return this.widgets.author_entry.text;
    }
  }

  public string project_version
  {
    owned get
    {
      return this.widgets.version_entry.text;
    }
  }

  public string project_license
  {
    owned get
    {
      string license = "None";
      Gtk.TreeIter iter;

      if (this.widgets.licenses_combo.get_active_iter (out iter))
      {
        this.licenses_list.get (iter, 0, out license);
      }
      return license;
    }
  }

  public string project_dir
  {
    owned get
    {
      return this.widgets.dir_chooser.get_filename ();
    }
  }  

  public string project_path
  {
    owned get
    {
      return this.widgets.path_entry.text;
    }
  }

  public string template
  {
    owned get;
    private set;
  }

  public string project_desc
  {
    set
    {
      this.widgets.desc_label.set_text (value);
    }
  }

  private void populate_icon_view ()
  {
    List<string> files = TemplateManager.get_instance ().get_list ();

    this.list_store = new Gtk.ListStore (ColumnType.NB_COLUMNS,
                                         typeof (Gdk.Pixbuf),
                                         typeof (string),
                                         typeof (string),
                                         typeof (string));
    this.widgets.icon_view.set_model (this.list_store);

    foreach (string filename in files)
    {
      Gtk.TreeIter iter;
      Template template = new Template (filename);

      this.list_store.append (out iter);
      this.list_store.set (iter, ColumnType.ICON, template.icon,
                                 ColumnType.NAME, template.name,
                                 ColumnType.DESCRIPTION, template.desc,
                                 ColumnType.PATH, template.path,
                                 -1);
    }
  }

  private void populate_license ()
  {
    this.licenses_list = new Gtk.ListStore (1, typeof (string));
    this.widgets.licenses_combo.set_model (this.licenses_list);

    Gtk.CellRendererText cell = new Gtk.CellRendererText ();
    this.widgets.licenses_combo.pack_start (cell, true);
    this.widgets.licenses_combo.set_attributes (cell, "text", 0);

    try
    {
      string filename;
      List<string> files = new List<string> ();
      Dir dir = Dir.open (Path.build_filename (Config.PACKAGE_DATA_DIR, "licenses"));

      while ((filename = dir.read_name ()) != null)
      {
        files.append (filename);
      }
      files.sort (strcmp);

      foreach (string file in files)
      {
        Gtk.TreeIter iter;

        this.licenses_list.append (out iter);
        this.licenses_list.set (iter, 0, file);
      }
    }
    catch (Error e)
    {
      debug (e.message);
    }
  }

  [CCode (instance_pos = -1)]
  protected void selection_changed (Gtk.IconView icon_view)
  {
    Gtk.TreeIter iter;

    weak List<Gtk.TreePath*> item = icon_view.get_selected_items ();

    if (item != null)
    {
      Gtk.TreePath* path = item.data;
      Gtk.TreePath path2 = new Gtk.TreePath.from_string (path->to_string ());
      this.list_store.get_iter (out iter, path2);

      string tmp;
      this.list_store.get (iter, ColumnType.PATH, out tmp);
      this.template = Path.build_filename (tmp, "template.xml");

      this.list_store.get (iter, ColumnType.DESCRIPTION, out tmp);
      this.project_desc = tmp;
      this.widgets.assistant.set_page_complete (this.widgets.page2, true);
    }
    else
    {
      this.widgets.assistant.set_page_complete (this.widgets.page2, false);
    }
  }

  [CCode (instance_pos = -1)]
  protected void project_path_change (Gtk.Widget sender)
  {
    this.widgets.path_entry.set_text (Path.build_filename (this.project_dir,
                                                           this.project_name) +  Path.DIR_SEPARATOR_S);
    if (this.project_name != "")
    {
      this.widgets.assistant.set_page_complete (this.widgets.page3, true);
    }
    else
    {
      this.widgets.assistant.set_page_complete (this.widgets.page3, false);
    }
  }

  construct
  {
    try
    {
      Gdk.Pixbuf icon = new Gdk.Pixbuf.from_file (Path.build_filename (Config.PACKAGE_PIXMAPS_DIR, "icone-32.xpm"));

      this.widgets.assistant.set_page_header_image (this.widgets.page1, icon);
      this.widgets.assistant.set_page_header_image (this.widgets.page2, icon);
      this.widgets.assistant.set_page_header_image (this.widgets.page3, icon);
    }
    catch (Error e)
    {
      debug (e.message);
    }

    this.populate_icon_view ();
    this.populate_license ();
    this.connect_signals ("valide_project_dialog_");
    this.widgets.assistant.set_page_complete (this.widgets.page1, true);
    this.widgets.assistant.apply += () => {
      this.apply ();
    };
    this.widgets.assistant.cancel += () => {
      this.widgets.assistant.destroy ();
    };
    this.widgets.assistant.close += () => {
      this.widgets.assistant.destroy ();
    };
  }

  public void show ()
  {
    this.widgets.assistant.show_all ();
  }

  /* Rewriting of the Gtk.Dialog.run function */
  private struct RunInfo
  {
    public MainLoop loop;
    public Gtk.ResponseType response_id;
    public bool destroyed;
  }

  private static void shutdown_loop (RunInfo ri)
  {
    if (ri.loop.is_running ())
    {
      ri.loop.quit ();
    }
  }

  [Callback]
  private static void apply_cb (Gtk.Widget sender, ref RunInfo ri)
  {
    ri.response_id = Gtk.ResponseType.APPLY;
    ProjectDialog.shutdown_loop (ri);
  }

  [Callback]
  private static void cancel_cb (Gtk.Widget sender, ref RunInfo ri)
  {
    ri.response_id = Gtk.ResponseType.CANCEL;
    ProjectDialog.shutdown_loop (ri);
  }

  [Callback]
  private static void close_cb (Gtk.Widget sender, ref RunInfo ri)
  {
    ri.response_id = Gtk.ResponseType.CLOSE;
    ProjectDialog.shutdown_loop (ri);
  }


  [Callback]
  private static void unmap_cb (Gtk.Widget sender, ref RunInfo ri)
  {
    ProjectDialog.shutdown_loop (ri);
  }

  [Callback]
  private static bool delete_cb (Gtk.Widget sender, ref RunInfo ri)
  {
    ProjectDialog.shutdown_loop (ri);
    return true;
  }

  [Callback]
  private static void destroy_cb (Gtk.Widget sender, ref RunInfo ri)
  {
    /* shutdown_loop will be called by run_unmap_handler */
    ri.destroyed = true;
  }

  public uint run ()
  {
    RunInfo ri = {null, Gtk.ResponseType.NONE, false};

    this.show ();
    Signal.connect (this.widgets.assistant, "apply", (Callback)apply_cb, (void*)(&ri));
    Signal.connect (this.widgets.assistant, "cancel", (Callback)cancel_cb, (void*)(&ri));
    Signal.connect (this.widgets.assistant, "close", (Callback)close_cb, (void*)(&ri));
    Signal.connect (this.widgets.assistant, "delete_event", (Callback)delete_cb, (void*)(&ri));
    Signal.connect (this.widgets.assistant, "unmap", (Callback)unmap_cb, (void*)(&ri));
    Signal.connect (this.widgets.assistant, "destroy", (Callback)destroy_cb, (void*)(&ri));

    ri.loop = new MainLoop (null, false);
    Gdk.threads_leave ();
    ri.loop.run ();
    Gdk.threads_enter ();
    ri.loop = null;
    if (!ri.destroyed)
    {
      SignalHandler.disconnect_by_func (this.widgets.assistant, (void*)apply_cb, (void*)(&ri));
      SignalHandler.disconnect_by_func (this.widgets.assistant, (void*)cancel_cb, (void*)(&ri));
      SignalHandler.disconnect_by_func (this.widgets.assistant, (void*)close_cb, (void*)(&ri));
      SignalHandler.disconnect_by_func (this.widgets.assistant, (void*)delete_cb, (void*)(&ri));
      SignalHandler.disconnect_by_func (this.widgets.assistant, (void*)unmap_cb, (void*)(&ri));
      SignalHandler.disconnect_by_func (this.widgets.assistant, (void*)destroy_cb, (void*)(&ri));
    }
    return ri.response_id;
  }
}

