/* project-manager.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <gege2061@redaction-developpez.com>
 */

public errordomain Valide.ProjectError
{
  NO_PROJECT,
  MAX_OPEN
}

public class Valide.ProjectManager : Gtk.ScrolledWindow
{
  enum Col
  {
    ICON,
    NAME,
    PATH,
    SHORT_PATH,
    NB
  }

  private static ProjectManager instance = null;
  private Project project;
  private List<Project> projects;
  private Gtk.TreeStore tree_store;
  private Gtk.TreeView tree_view;
  private weak Project _current;
  private Gtk.Menu popup;
  private ProjectDialog project_dialog;

  private string copy_template (Template template,
                                string path, string name, string author,
                                string version, string license) throws Error
  {
    string filename = null;
    string contents = null;

    // Transform XML
    Xml.Doc* xml = Xml.Parser.parse_file (template.filename);
    Xml.XPathContext xpath = new Xml.XPathContext (xml);
    Xml.XPathObject* xpath_obj = xpath.eval ("/template");
    xpath_obj->nodesetval->item (0)->name = "project";

    xpath_obj = xpath.eval ("/project/description");
    xpath_obj->nodesetval->item (0)->unlink ();
    xpath_obj = xpath.eval ("/project/icon");
    xpath_obj->nodesetval->item (0)->unlink ();
    xpath_obj = xpath.eval ("/project/name");
    xpath_obj->nodesetval->item (0)->set_content (name);
    xpath_obj = xpath.eval ("/project");
    xpath_obj->nodesetval->item (0)->new_text_child (null, "author", author);
    xpath_obj->nodesetval->item (0)->new_text_child (null, "version", version);
    xpath_obj->nodesetval->item (0)->new_text_child (null, "license", license);

    // Copy license
    if (license != "None")
    {
      FileUtils.get_contents (Path.build_filename (Config.PACKAGE_DATA_DIR, "licenses", license),
                              out contents, null);
      FileUtils.set_contents (Path.build_filename (path, "COPYING"),
                              contents, -1);
      xpath_obj = xpath.eval ("/project/files");
      xpath_obj->nodesetval->item (0)->new_text_child (null, "file", "COPYING");
    }

    // Copy xml
    filename = Path.build_filename (path, name + "." + Project.FILE_EXT[0]);
    xml->save_file (filename);

    // Copy source files
    foreach (string file in template.files)
    {
      File src = File.new_for_path (file);
      File dst = File.new_for_path (Path.build_filename (path, Path.get_basename (file)));

      src.copy (dst, FileCopyFlags.OVERWRITE, null, null);
    }

    return filename;
  }

  public Project current
  {
    get
    {
      return this._current;
    }
    set
    {
      this._current = value;
      this.refresh ();
    }
  }

  [Callback]
  private void file_select (Gtk.TreeView sender, Gtk.TreePath path)
  {
    Gtk.TreeIter iter;
    string filename = null;

    this.tree_store.get_iter (out iter, path);
    this.tree_store.get (iter, Col.PATH, out filename);
    if (FileUtils.test (filename, FileTest.IS_REGULAR))
    {
      try
      {
        this.current.open_file (filename);
      }
      catch (Error e)
      {
        debug (e.message);
      }
    }
    else
    {
      // @todo Active project
      // @todo Unexpand other projects
      if (this.tree_view.row_is_expanded (path))
      {
        this.tree_view.collapse_row (path);
      }
      else
      {
        this.tree_view.expand_row (path, false);
      }
    }
  }

  [Callback]
  private bool button_press (Gtk.TreeView sender, Gdk.EventButton event)
  {
    if (event.type == Gdk.EventType.BUTTON_PRESS && event.button == 3)
    {
      this.popup.popup (null, null, null, event.button, event.time);
      return true;
    }
    return false;
  }

  private void setup_ui ()
  {
    bool active = false;
    UIManager ui = UIManager.get_instance ();

    if (this.current != null)
    {
      active = true;
    }
    ui.get_action ("project-add").sensitive = active;
    ui.get_action ("project-remove").sensitive = active;
    ui.get_action ("project-options").sensitive = active;
    ui.get_action ("project-compile").sensitive = active;
    ui.get_action ("project-execute").sensitive = active;
    ui.get_action ("project-abort").sensitive = active;
    ui.get_action ("project-close").sensitive = active;
    // @TODO
    //((Gtk.ToggleAction)ui.get_action ("view-side-panel")).active = active;
    this.sensitive = active;
    ExecutableManager.get_instance ().sensitive = active;
  }

  /* @TODO Save all expanded rows (childrens of root) */
  private void refresh ()
  {
    bool expanded;
    Gtk.TreePath path = new Gtk.TreePath ();
    Gtk.TreeIter iter;

    /* Save expanded rows */
    expanded = this.tree_store.get_iter_first (out iter);
    if (expanded == true)
    {
      path = this.tree_store.get_path (iter);
      expanded &= this.tree_view.row_is_expanded (path);
    }

    this.tree_store.clear ();
    foreach (Project p in projects)
    {
      string title = null;
      Gtk.TreeIter root;

      this.tree_store.append (out root, null);

      if (p == this.current)
      {
        this.tree_store.set (root, Col.NAME, "<b>" + p.name + "</b>", -1);
      }
      else
      {
        this.tree_store.set (root, Col.NAME, p.name, -1);
      }

      HashTable<string, Gtk.TreeIter*> dir_iters;

      dir_iters = new HashTable<string, Gtk.TreeIter*> (str_hash, str_equal);
      foreach (string file in p.files)
      {
        string dir = Path.get_dirname (file);
        string home = Environment.get_home_dir ();
        string filename = Path.get_basename (file);
        string filepath;

        if (Path.is_absolute (file))
        {
          filepath = file;
        }
        else
        {
          filepath = Path.build_filename (p.path, file);
        }
        if (dir.has_prefix (home))
        {
          dir = "~" + dir.substring (home.size (),
                                     dir.size () - home.size ());
        }

        Gtk.TreeIter* dir_iter;
        if (dir == ".")
        {
          dir_iter = &root;
        }
        else
        {
          dir_iter = dir_iters.lookup (dir);

          if (dir_iter == null)
          {
            Gtk.TreeIter tmp;

            this.tree_store.append (out tmp, root);
            dir_iter = (Gtk.TreeIter*)tmp.copy ();
            this.tree_store.set (*dir_iter, Col.ICON,
                                 Utils.get_pixbuf_for_theme (Gtk.STOCK_DIRECTORY, Gtk.IconSize.MENU),
                                 Col.NAME, dir, -1);
            dir_iters.insert (dir, dir_iter);
          }
        }
        this.tree_store.append (out iter, *dir_iter);
        this.tree_store.set (iter, Col.NAME, filename, Col.ICON,
                             /* @todo Use mime type corresponding icon */
                             Utils.get_pixbuf_for_theme (Gtk.STOCK_FILE, Gtk.IconSize.MENU),
                             Col.PATH, filepath, Col.SHORT_PATH, file);
      }
    }
    this.setup_ui ();

    /* Restore expanded rows */
    if (this.sensitive && expanded)
    {
      this.tree_view.expand_row (path, true);
    }
  }

  private ProjectManager ()
  {
    this.width_request = 200;
  }

  construct
  {
  	this.set_policy (Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC);
    /* create some nice padding */
    this.set_shadow_type (Gtk.ShadowType.IN);
    this.set_border_width (2);

  	this.tree_view = new Gtk.TreeView ();
    this.tree_view.headers_visible = false;
  	this.add (this.tree_view);

    this.tree_store = new Gtk.TreeStore (Col.NB, typeof (Gdk.Pixbuf),
                                         typeof (string), typeof(string),
                                         typeof (string));
    /* @todo Use sorting function for show directories before */
    this.tree_store.set_sort_column_id (Col.NAME, Gtk.SortType.ASCENDING);
    this.tree_view.set_model (this.tree_store);

    Gtk.TreeViewColumn col = new Gtk.TreeViewColumn ();
    col.set_title(_("Projects"));
    {
      Gtk.CellRendererPixbuf renderer = new Gtk.CellRendererPixbuf ();
      col.pack_start (renderer, false);
      col.set_attributes (renderer, "pixbuf", Col.ICON);
    }
    {
      Gtk.CellRendererText renderer = new Gtk.CellRendererText ();
      col.pack_start (renderer, false);
      col.set_attributes (renderer, "markup", Col.NAME);
    }
    this.tree_view.append_column (col);

    this.tree_view.row_activated += this.file_select;
    /* Add popup */
    UIManager ui_manager = UIManager.get_instance ();
    this.popup = (Gtk.Menu)ui_manager.get_widget ("/project-popup");
    this.tree_view.button_press_event += this.button_press;

    this.setup_ui ();
  }

  public override void destroy ()
  {
    this.instance = null;
    base.destroy ();
  }

  public static ProjectManager get_instance ()
  {
    if (ProjectManager.instance == null)
    {
      ProjectManager.instance = new ProjectManager ();
    }
    return ProjectManager.instance;
  }

  public bool create ()
  {
    bool created = false;
    ProjectDialog dialog = new ProjectDialog ();

    if (dialog.run () == Gtk.ResponseType.APPLY)
    {
      string path = dialog.project_dir;

      if (!FileUtils.test (path, FileTest.EXISTS))
      {
        Gtk.Dialog msg = new Gtk.Dialog.with_buttons (_("New project"), null,
                                                      Gtk.DialogFlags.MODAL,
                                                      Gtk.STOCK_NO, Gtk.ResponseType.NO,
                                                      Gtk.STOCK_YES, Gtk.ResponseType.YES,
                                                      null);

        weak Gtk.Box main_box = (Gtk.Box)msg.vbox;
        main_box.pack_start (new Gtk.Label (_("Projects directory doesn't exist. Would you create it?")),
                             true, true, 10);
        msg.show_all ();
        if (msg.run () == Gtk.ResponseType.YES)
        {
          DirUtils.create_with_parents (path, 0755);
          created = true;
        }
        msg.destroy ();
      }
      else
      {
        created = true;
      }

      if (created == true)
      {
        path = dialog.project_path;
        if (FileUtils.test (path, FileTest.EXISTS))
        {
          Gtk.Dialog msg = new Gtk.Dialog.with_buttons (_("New project"), null,
                                                        Gtk.DialogFlags.MODAL,
                                                        Gtk.STOCK_NO, Gtk.ResponseType.NO,
                                                        Gtk.STOCK_YES, Gtk.ResponseType.YES,
                                                        null);

          weak Gtk.Box main_box = (Gtk.Box)msg.vbox;
          main_box.pack_start (new Gtk.Label (_("Directory already exist. Would you overwrite it?")),
                               true, true, 10);
          msg.show_all ();
          if (msg.run () == Gtk.ResponseType.NO)
          {
            created = false;
          }
          msg.destroy ();
        }
        else
        {
          DirUtils.create (path, 0755);
        }
      }

      if (created == true)
      {
        try
        {
          string prj = this.copy_template (new Template (dialog.template),
                                           path, dialog.project_name,
                                           dialog.project_author,
                                           dialog.project_version,
                                           dialog.project_license);
          this.open (prj);
        }
        catch (Error e)
        {
          warning (e.message);
          created = false;
        }
      }
    }
    dialog.destroy ();
    return created;
  }

  public bool open (string? filename = null) throws ProjectError
  {
    bool opened = false;

    if (current == null)
    {
      if (filename == null)
      {
        Gtk.FileChooserDialog dialog = new Gtk.FileChooserDialog (_("Open project"), null,
                                                                  Gtk.FileChooserAction.OPEN,
                                                                  Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                                                  Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT,
                                                                  null);

        /* Add filter */
        Gtk.FileFilter filter = new Gtk.FileFilter ();
        filter.set_name (_("Val(a)IDE project file"));
        foreach (string ext in Project.FILE_EXT)
        {
          filter.add_pattern ("*." + ext);
        }
        dialog.add_filter (filter);

        if (dialog.run () == Gtk.ResponseType.ACCEPT)
        {
          try
          {
            filename =  dialog.get_filename ();
            Gtk.RecentManager recent_manager = Gtk.RecentManager.get_default ();
            recent_manager.add_item (dialog.get_filename ());
          }
          catch (Error e)
          {
            warning (e.message);
          }
        }
        dialog.destroy ();
      }
      if (filename != null)
      {
        if (FileUtils.test (filename, FileTest.EXISTS))
        {
          Project p = new Project (filename);

          RecentManagerUI.get_instance ().add_file (filename);
          p.closed += this.close;
          p.added_file += this.refresh;
          p.removed_file += this.refresh;
          p.options_changed += this.refresh;
          this.projects.append (p);
          this.current = p;
          this.refresh ();
          ((Gtk.ToggleAction)UIManager.get_instance ()
            .get_action ("view-side-panel")).set_active (true);

          opened = true;
        }
        else
        {
          warning (_("The project file doesn't exists!"));
        }
      }
    }
    else
    {
      throw new ProjectError.MAX_OPEN (_("Maximum projects open!"));
    }
    return opened;
  }

  public void close (Project sender)
  {
    this.projects.remove (sender);
    this.current = null;
  }

  public void open_selected_file ()
  {
    Gtk.TreeIter iter;
    Gtk.TreeSelection selection = this.tree_view.get_selection ();

    if (selection.get_selected (null, out iter))
    {
      Gtk.TreePath path = this.tree_store.get_path (iter);

      this.file_select (this.tree_view, path);
    }
  }

  public void remove_selected_file ()
  {

    Gtk.TreeIter iter;
    Gtk.TreeSelection selection = this.tree_view.get_selection ();

    if (selection.get_selected (null, out iter))
    {
      Gtk.TreePath path = this.tree_store.get_path (iter);

      if (path.get_depth () > 1)
      {
        string filename;
        List<string> files = null;

        this.tree_store.get (iter, Col.SHORT_PATH, out filename);
        files.append (filename);
        this.current.remove_file (files.copy ());
      }
    }
  }

  /**
   * Test if a filename is a project file.
   *
   * @param filename: a filename.
   *
   * @return true if a filename is a project file.
   */
  public bool accept_file (string filename)
  {
    bool accept = false;
    string ext = Utils.get_extension (filename);

    foreach (string e in Project.FILE_EXT)
    {
      if (e == ext)
      {
        accept = true;
        break;
      }
    }
    return accept;
  }
}

