/* item.vala
 *
 * Copyright (C) 2008  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   Nicolas Joseph <gege2061@redaction-developpez.com>
 */

/*
 *  Vala symbol completion library
 *  
 *  Copyright (C) 2008 - Andrea Del Signore <sejerpz@tin.it>
 */

public class Completion.Item : Object
{
  public string name;
  public string type_name = null;
  public string info;
  public int line = 0;

  private string data_type_to_string (Vala.DataType type)
  {
    string result;

    result = type.to_qualified_string ();
    if (result == null)
    {
      result = "void";
    }

    Gee.List<Vala.DataType> type_args = type.get_type_arguments ();
    if (type_args.size > 0)
    {
      result += "&lt;";
      bool first = true;
      foreach (Vala.DataType type_arg in type_args)
      {
        if (!first)
        {
          result += ",";
        }
        else
        {
          first = false;
        }
        if (!type_arg.value_owned)
        {
          result += "weak ";
        }
        result += this.data_type_to_string (type_arg);
      }
      result += "&gt;";
    }

    if (type.nullable && !result.has_suffix ("?") && !result.has_suffix("*")) 
    {
      result += "?";
    }
    if (type.is_dynamic)
    {
      result = "dynamic " + result;
    }

    return result;
  }

  private string formal_parameters_to_string (Gee.List<Vala.FormalParameter> parameters)
  {
    string params = "";
    string param_sep = " ";
    if (parameters.size > 2)
    {
      param_sep = "\n\t\t";
    }

    foreach (Vala.FormalParameter parameter in parameters)
    {
      string direction = "";
      string default_expr = "";
      string parameter_type = "";

      if (parameter.ellipsis)
      {
        params = "%s, %s".printf (params, "...");
      }
      else
      {
        if (parameter.direction == Vala.ParameterDirection.OUT)
        {
          direction = "out ";
        }
        else if (parameter.direction == Vala.ParameterDirection.REF)
        {
          direction = "ref ";
        }
        if (parameter.default_expression != null)
        {
          default_expr = " = " + parameter.default_expression.to_string ();
        }
        if (parameter.parameter_type != null)
        {
          parameter_type = this.data_type_to_string (parameter.parameter_type);
        }
        else
        {
          parameter_type = "<unkown>";
        }
        params = "%s,%s%s%s %s%s".printf (params, param_sep, direction, parameter_type, parameter.name, default_expr);
      }
    }
    if (params != "")
    {
      params = params.substring (2, params.length - 2);
    }
    return params;
  }

  public Item (string name)
  {
    this.name = name;
    this.info = name;
  }

  public Item.with_method (Vala.Method item)
  {
    this.name = item.name;
    this.line = item.source_reference.first_line;
    
    if (name.has_prefix ("new"))
    {
      name = name.substring (3, name.length - 3);
      if (name == "")
      {
        name = item.parent_symbol.name;
      }
      else if (name.has_prefix ("."))
      {
        name = name.substring (1, name.length - 1);
      }
    }

    int param_count = item.get_parameters ().size;
    string params = this.formal_parameters_to_string (item.get_parameters ());

    this.info = "Method: %s\n\n%s%s<b>%s</b> (%s%s)".printf (
        name,
        this.data_type_to_string (item.return_type),
        (param_count > 2 ? "\n" : " "),
        name, 
        (param_count > 2 ? "\n" : ""),
        params);
  }

  public Item.with_field (Vala.Field item)
  {
    this.name = item.name;
    this.line = item.source_reference.first_line;
    
    string default_expr = "";
    if (item.initializer != null)
    {
      default_expr = " = " + item.initializer.to_string ();
    }
    this.info = "Field: %s\n\n%s <b>%s</b>%s".printf (
        name,
        this.data_type_to_string (item.field_type),
        name, 
        default_expr);
  }

  public Item.with_property (Vala.Property item)
  {
    this.name = item.name;
    string default_expr = "";
    if (item.default_expression != null)
    {
      default_expr = " = " + item.default_expression.to_string ();
    }
    this.info = "Property: %s\n\n%s <b>%s</b>%s".printf (
        name,
        this.data_type_to_string (item.property_type),
        name, 
        default_expr);
  }

  public Item.with_struct (Vala.Struct item)
  {
    this.name = item.name;
    this.info = "Struct: %s".printf (item.name);
    this.line = item.source_reference.first_line;
  }

  public Item.with_class (Vala.Class item)
  {
    this.name = item.name;
    this.info = "Class: %s".printf (item.name);
    this.line = item.source_reference.first_line;
  }

  public Item.with_interface (Vala.Interface item)
  {
    this.name = item.name;
    this.info = "Interface: %s".printf (item.name);
    this.line = item.source_reference.first_line;
  }

  public Item.with_signal (Vala.Signal item)
  {
    this.name = item.name;
    this.info = "Signal: %s".printf (item.name);
    this.line = item.source_reference.first_line;
    int param_count = item.get_parameters ().size;
    string params = this.formal_parameters_to_string (item.get_parameters ());

    this.info = "Signal: %s\n\n%s%s<b>%s</b> (%s%s)".printf (
        name,
        this.data_type_to_string (item.return_type),
        (param_count > 2 ? "\n" : " "),
        name, 
        (param_count > 2 ? "\n" : ""),
        params);
  }
}

