/* documentwords-provider.vala
 *
 * Copyright (C) 2008  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *  Nicolas Joseph <gege2061@redaction-developpez.com>
 */

/*
 * This code is re-writing in the vala language of GtkSourceCompletion
 *
 * Copyright (C) 2007 - Jess Barbero
 *
 * rev 163
 */

public class SourceCompletion.DocumentwordsProvider : Object, Provider
{
  private bool is_completing;
  private HashTable<string, int> current_words;
  private List<Proposal> data_list;
  private string cleaned_word;
  private Gdk.Pixbuf icon;
  private int count;
  private Gtk.TextIter start_iter;
  private const string PROVIDER_NAME = "GscDocumentwordsProvider";
  private const string ICON_FILE = "default.png";

  public enum SortType
  {
    NONE,
    BY_LENGTH
  }

  public Gtk.TextView view
  {
    get;
    construct;
  }

  public SortType sort_type
  {
    get;
    set;
  }

  public bool pred_is_separator (unichar ch)
  {
    return Utils.char_is_separator(ch);
  }

  public static int utf8_len_compare (void* a, void* b)
  {
    long lena,lenb;
    lena = ((Proposal)a).label.len ();
    lenb = ((Proposal)b).label.len ();
    if (lena == lenb)
      return 0;
    else if (lena < lenb)
      return -1;
    else
      return 1;
  }

  private HashTable get_all_words (Gtk.TextBuffer buffer)
  {
    Gtk.TextIter start_iter;
    Gtk.TextIter prev_iter;
    string word;
    HashTable<string, int> result = new HashTable<string, int> (str_hash, str_equal);

    buffer.get_start_iter (out start_iter);
    prev_iter = start_iter;
    while (start_iter.forward_find_char (pred_is_separator, null))
    {
      if (start_iter.compare (prev_iter) != 0)
      {
        word = prev_iter.get_text (start_iter);

        if (word.len () > 0)
          result.insert (word, 0);
      }
      prev_iter = start_iter;
      prev_iter.forward_char ();
    }

    if (!prev_iter.is_end ())
    {
      buffer.get_end_iter (out start_iter);
      if (start_iter.compare (prev_iter) != 0)
      {
        word = prev_iter.get_text (start_iter);
        if (word.len () > 0)
          result.insert (word, 0);
      }
      prev_iter = start_iter;
      prev_iter.forward_char ();
    }
    return result;
  }

  private void clean_current_words ()
  {
    // Clean the previous data
    if (this.current_words != null)
    {
      this.current_words = null;
    }
  }

  /*
   * Check the proposals hash and inserts the completion proposal into the final list
   */
  private void gh_add_key_to_list (void* key, void* val)
  {
    if (this.count >= 500)
    {
      return;
    }
    Proposal data;
    if (Utils.is_valid_word (this.cleaned_word, (string)key))
    {
      this.count++;
      data = new Proposal ((string)key, null, this.icon);
      this.data_list.append (data);
    }
  }

  private void sort_completion_list (List data_list)
  {
    switch (this.sort_type)
    {
      case SortType.BY_LENGTH:
      {
        data_list.sort (utf8_len_compare);
        break;
      }
      default: 
        break;
    }
  }

  public virtual string get_name ()
  {
    return this.PROVIDER_NAME;
  }

  public virtual List? get_proposals (Trigger trigger)
  {
    Gtk.TextIter tmp;
    string current_word = Utils.get_last_word_and_iter (this.view, out this.start_iter, out tmp);
    this.cleaned_word = Utils.clear_word (current_word);
  
    if (!this.is_completing)
    {
      Gtk.TextBuffer text_buffer = this.view.get_buffer ();
      this.current_words = this.get_all_words (text_buffer);
    }

    this.data_list = null;
    this.count = 0;
    this.current_words.for_each (gh_add_key_to_list);
    this.cleaned_word = null;

    if (this.data_list != null)
    {
      this.is_completing = true;
      this.sort_completion_list (this.data_list);
    }
    else
    {
      this.clean_current_words ();
      this.is_completing = false;
    }

    /* GscManager frees this list and data */
    return this.data_list.copy ();
  }

  public virtual void finish ()
  {
    /*Clean current word list*/
    this.clean_current_words ();

    this.is_completing = false;
  }

  construct
  {
    this.current_words = null;
    this.is_completing = false;
    this.count = 0;
    this.cleaned_word = null;
    try
    {
      this.icon = new Gdk.Pixbuf.from_file (Path.build_filename (Config.PACKAGE_PLUGIN_DIR, "completion", "pixmaps", this.ICON_FILE));
    }
    catch (Error e)
    {
      debug (e.message);
    }
    this.sort_type = SortType.BY_LENGTH;
  }

  public DocumentwordsProvider (Gtk.TextView view)
  {
    this.view = view;
  }
}

