/* popup.vala
 *
 * Copyright (C) 2008  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <gege2061@redaction-developpez.com>
 */

/*
 * This code is re-writing in the vala language of GtkSourceCompletion
 *
 * Copyright (C) 2007 - Chuchiperriman
 *
 * rev 163
 */

public class SourceCompletion.Popup : Gtk.Window
{
  private const int WINDOW_WIDTH=350;
  private const int WINDOW_HEIGHT=200;
  private bool destroy_has_run;
  private HashTable<string, Gtk.Widget> trees;
  private Gtk.Notebook notebook;
  private Gtk.ToggleButton info_button;
  private Gtk.Image next_page_icon;
  private Gtk.Label tab_label;
  private Gtk.Entry filter;
  private Info info_window;
  private Gtk.TextView view;
  private FilterType filter_type;
  private Gtk.HBox bottom_bar;
  private KeyDef[] keys;

  public signal void proposal_selected (Proposal proposal);
  public signal void display_info (Proposal proposal);

  public enum PositionType
  {
    CURSOR,
    CENTER_SCREEN,
    CENTER_PARENT
  }

  public enum FilterType
  {
    NONE,
    TREE,
    TREE_HIDDEN
  }

  public enum KeysType
  {
    INFO,
    PAGE_NEXT,
    PAGE_PREV,
    LAST
  }

  private struct KeyDef
  {
    public uint key;
    public Gdk.ModifierType mods;
  }

  private Tree get_current_tree ()
  {
    int page = this.notebook.get_current_page ();
    Tree tree = (Tree)notebook.get_nth_page (page);
    return tree;
  }

  private Tree get_tree_by_name (string tree_name)
  {
    Tree tree = (Tree)this.trees.lookup (tree_name);

    if (tree == null)
    {
      // We create the new trees
      Tree completion_tree = new Tree (); 
      this.trees.insert (tree_name, completion_tree);
      Gtk.Label label = new Gtk.Label (tree_name);
      notebook.append_page (completion_tree, label);
      completion_tree.show_all ();
      completion_tree.proposal_selected += proposal_selected_cb;
      completion_tree.selection_changed += selection_changed_cd;
    }
    return tree;
  }

  private void show_completion_info ()
  {
    Proposal proposal;
    this.get_selected_proposal (out proposal);
    if (proposal != null)
    {
      this.display_info (proposal);
    }
    int y, x, sw, sh;
    this.get_position (out x, out y);
    sw = Gdk.Screen.width ();
    sh = Gdk.Screen.height ();
    x += this.WINDOW_WIDTH;

    if (x + this.WINDOW_WIDTH >= sw)
    {
      x -= (this.WINDOW_WIDTH * 2);
    }
    this.info_window.move (x, y);
    this.info_window.set_transient_for (this.get_transient_for ());
    this.info_window.show ();
  }

  [Callback]
  private void info_toggled_cb (Gtk.ToggleButton widget)
  {
    if (widget.get_active ())
    {
      this.show_completion_info ();
    }
    else
    {
      this.info_window.hide ();
    }
  }

  [Callback]
  private void proposal_selected_cb (Tree tree, Proposal proposal)
  {
    this.proposal_selected (proposal);
  }

  [Callback]
  private void selection_changed_cd (Tree tree, Proposal proposal)
  {
    if (this.info_window.visible)
    {
      this.show_completion_info ();
    }
  }

  [Callback]
  private bool focus_out_event_cb (Gtk.Widget widget, Gdk.EventFocus event)
  {
    // Only Close if the filter is visible
    if (((Gtk.Widget)this.filter).visible)
    {
      this.hide ();
    }
    return false;
  }

  [Callback]
  private void switch_page_cb (Gtk.Notebook notebook, void* page, uint page_num)
  {
    Tree tree = (Tree)notebook.get_nth_page ((int)page_num);
    string label_text = notebook.get_tab_label_text (tree);

    this.tab_label.set_label (label_text);
  }

  private void update_pages_visibility ()
  {
    int pages = this.notebook.get_n_pages ();
    uint num_pages_with_data = 0;
    bool first_set = false;
    Tree tree;

    for(int i = 0; i < pages; i++)
    {
      tree = (Tree)this.notebook.get_nth_page (i);
      if (tree.get_num_proposals () > 0)
      {
        if (!first_set)
        {
          this.notebook.set_current_page (i);
          first_set = true;
        }
        num_pages_with_data++;
      }
    }
    if (num_pages_with_data > 1)
      this.next_page_icon.show ();
    else
      this.next_page_icon.hide ();
  }

  [Callback]
  private bool filter_key_press_event_cb (Gtk.Widget view, Gdk.EventKey event)
  {
    return this.manage_key (event);
  }

  public bool manage_key (Gdk.EventKey event)
  {
    bool catched = false;
    bool ret = false;

    switch (event.keyval)
    {
      case Gdk.KeySyms.Escape:
        this.hide ();
        catched = true;
        ret = true;
      break;
      case Gdk.KeySyms.Down:
        ret = this.select_next (1);
        catched = true;
      break;
      case Gdk.KeySyms.Page_Down:
        ret = this.select_next (5);
        catched = true;
      break;
      case Gdk.KeySyms.Up:
        ret = this.select_previous (1);
        if (!ret)
          ret = this.select_first ();
        catched = true;
      break;
      case Gdk.KeySyms.Page_Up:
        ret = this.select_previous (5);
        if (!ret)
          ret = this.select_last ();
        catched = true;
      break;
      case Gdk.KeySyms.Home:
        ret = this.select_first ();
        catched = true;
      break;
      case Gdk.KeySyms.End:
        ret = this.select_last ();
        catched = true;
      break;
      case Gdk.KeySyms.Return:
      case Gdk.KeySyms.Tab:
        if ((event.state & Gdk.ModifierType.CONTROL_MASK) == Gdk.ModifierType.CONTROL_MASK)
        {
          break;
        }
        Proposal prop = null;
        if (this.get_selected_proposal (out prop))
        {
          this.proposal_selected (prop);
          this.hide ();
        }
        else
        {
          this.hide ();
        }
        ret = true;
        catched = true;
      break;
    }
    if (!catched)
    {
      if (Utils.compare_keys (this.keys[(int)KeysType.INFO].key,
                              this.keys[(int)KeysType.INFO].mods,
                              event))
      {
        this.toggle_proposal_info ();
        ret = true;
      }
      else if (Utils.compare_keys (this.keys[(int)KeysType.PAGE_NEXT].key,
                                   this.keys[(int)KeysType.PAGE_NEXT].mods,
                                   event))
      {
        this.page_next ();
        ret = true;
      }
      else if (Utils.compare_keys (this.keys[(int)KeysType.PAGE_PREV].key,
                                   this.keys[(int)KeysType.PAGE_PREV].mods,
                                   event))
      {
        this.page_previous ();
        ret = true;
      }
    }
    return ret;
  }

  [Callback]
  private void filter_changed_cb (Gtk.Entry sender)
  {
    int pages = this.notebook.get_n_pages ();

    for (int i = 0; i < pages; i++)
    {
      Tree tree = (Tree)this.notebook.get_nth_page (i);
      tree.filter (sender.get_text ());
    }
    this.select_first ();
  }

  public void show_or_update ()
  {
    /*TODO This function called only if the widget has not been shown previously*/

    /*Only show the popup, the positions is set before this function*/

    /*
    this.set_transient_for (parent);
    */

    this.update_pages_visibility ();

    if (!this.visible)
    {
      this.info_button.set_active (false);
      base.show ();
    }

    this.filter.set_text ("");

    /*Filter*/
    switch (this.filter_type)
    {
      case FilterType.TREE_HIDDEN:
      case FilterType.NONE:
        this.filter.hide ();
      break;
      default:
        this.filter.show ();
      break;
    }
    this.present ();
    this.activate_focus ();
    this.filter.grab_focus ();
    this.get_current_tree ().select_first ();
  }

  public override void realize ()
  {
    this.set_border_width (1);
    this.set_size_request (this.WINDOW_WIDTH, this.WINDOW_HEIGHT);
    this.set_resizable (true);
    base.realize ();
  }

  [Callback]
  private bool delete_event_cb (Gtk.Widget widget, Gdk.Event  event)
  {
    // Prevent the alt+F4 keys
    return true;
  }

  public override void destroy ()
  {
    if (!this.destroy_has_run)
    {
      this.clear ();
      this.destroy_has_run = true;
    }
    base.destroy ();
  }

  construct
  {
    this.keys = new KeyDef[(int)KeysType.LAST];
    this.set_type_hint (Gdk.WindowTypeHint.NORMAL);
    this.set_focus_on_map (false);
    this.set_size_request (this.WINDOW_WIDTH, this.WINDOW_HEIGHT);
    this.set_decorated (false);

    this.destroy_has_run = false;
    this.trees = new HashTable<string, Gtk.Widget> (str_hash, str_equal);

    Tree completion_tree = new Tree ();
    completion_tree.can_focus = false;

    this.trees.insert (Proposal.DEFAULT_PAGE, completion_tree);
	
    this.notebook = new Gtk.Notebook ();
    this.notebook.can_focus = false;
    this.notebook.set_show_tabs (false);
    this.notebook.set_show_border (false);

    Gtk.Label default_label = new Gtk.Label (Proposal.DEFAULT_PAGE);
    this.notebook.append_page (completion_tree, default_label);

    // Icon list
    Gtk.Image info_icon = new Gtk.Image.from_stock (Gtk.STOCK_INFO, Gtk.IconSize.SMALL_TOOLBAR);
    info_icon.set_tooltip_text (_("Show Proposal Info"));
    this.info_button = new Gtk.ToggleButton ();
    this.info_button.can_focus = false;
    this.info_button.set_focus_on_click (false);
    this.info_button.add (info_icon);
    this.info_button.toggled += info_toggled_cb;

    this.bottom_bar = new Gtk.HBox (false, 1);
    this.bottom_bar.pack_start (info_button, false, false, 0);

    // Next page icon
    this.next_page_icon = new Gtk.Image.from_stock(Gtk.STOCK_GO_FORWARD, Gtk.IconSize.SMALL_TOOLBAR);
    this.bottom_bar.pack_end (this.next_page_icon, false, false, 1);

    // Page label
    this.tab_label = new Gtk.Label (Proposal.DEFAULT_PAGE);
    this.bottom_bar.pack_end (this.tab_label, false, true, 10);

    // Filter entry
    this.filter = new Gtk.Entry ();

    Gtk.VBox vbox = new Gtk.VBox (false, 1);
    vbox.pack_start (this.filter, false, false, 0);

    vbox.pack_start (this.notebook, true, true, 0);

    vbox.pack_end (this.bottom_bar, false, false, 0);

    this.add (vbox);
    vbox.show_all ();

    // Info window
    this.info_window = new Info ();
    this.info_window.can_focus = false;
    this.info_window.accept_focus = false;

    // Connect signals
    completion_tree.proposal_selected += proposal_selected_cb;
    completion_tree.selection_changed += selection_changed_cd;
    this.notebook.switch_page += switch_page_cb;
    this.filter.changed += filter_changed_cb;
    this.filter.key_release_event += filter_key_press_event_cb;
    this.delete_event += delete_event_cb;
    this.focus_out_event += focus_out_event_cb;
  }

  public Popup ()
  {
    this.type = Gtk.WindowType.POPUP;
  }

  public void add_data (Proposal data)
  {
    Tree tree = this.get_tree_by_name (data.page_name);

    tree.add_data (data);
  }

  public void clear ()
  {
    int pages = this.notebook.get_n_pages ();
    Tree tree;

    for (int i = 0; i < pages; i++)
    {
      tree = (Tree)this.notebook.get_nth_page (i);
      tree.clear ();
    }
  }

  public bool select_first ()
  {
    return this.get_current_tree ().select_first ();
  }

  public bool select_last ()
  {
    return this.get_current_tree ().select_last ();
  }

  public bool select_previous (int rows)
  {
    return this.get_current_tree ().select_previous (rows);
  }

  public bool select_next (int rows)
  {
    return this.get_current_tree ().select_next(rows);
  }

  public bool get_selected_proposal (out Proposal proposal)
  {
    return this.get_current_tree ().get_selected_proposal (out proposal);
  }

  public bool has_proposals ()
  {
    int pages = this.notebook.get_n_pages ();
    Tree tree;
    for (int i = 0; i < pages; i++)
    {
      tree = (Tree)this.notebook.get_nth_page (i);
      if (tree.get_num_proposals () > 0)
        return true;
    }
    return false;
  }

  public void toggle_proposal_info ()
  {
    this.info_button.set_active (!this.info_button.get_active ());
  }

  public void page_next ()
  {
    Tree tree;
    int pages = this.notebook.get_n_pages ();
    int page = this.notebook.get_current_page ();
    int original_page = page;

    do
    {
      page++;
      if (page == pages)
        page = 0;

      tree = (Tree)this.notebook.get_nth_page (page);
      if (tree.get_num_proposals () > 0)
      {
        this.notebook.set_current_page (page);
        tree.select_first ();
        break;
      }
    } while (page != original_page);

    if (page != original_page)
    {
      if (this.info_window.visible)
      {
        this.show_completion_info ();
      }
    }
  }

  public void page_previous ()
  {
    Tree tree;
    int pages = this.notebook.get_n_pages ();
    int page = this.notebook.get_current_page ();
    int original_page = page;

    do
    {
      page--;
      if (page < pages)
        page = pages - 1;

      tree = (Tree)this.notebook.get_nth_page (page);
      if (tree.get_num_proposals () > 0)
      {
        this.notebook.set_current_page (page);
        tree.select_first ();
        break;
      }
    } while (page != original_page);

    if (page != original_page)
    {
      if (this.info_window.visible)
      {
        this.show_completion_info ();
      }
    }
  }

  public void set_current_info (string? info)
  {
    if (info != null)
    {
      this.info_window.set_markup (info);
    }
    else
    {
      this.info_window.set_markup (_("There is no info for the current proposal"));
    }
  }

  public int get_num_active_pages ()
  {
    int pages = this.notebook.get_n_pages ();
    int num_pages_with_data = 0;
    Tree tree;
    for (int i = 0; i < pages; i++)
    {
      tree = (Tree)this.notebook.get_nth_page (i);
      if (tree.get_num_proposals () > 0)
      {
        ++num_pages_with_data;
      }
    }
    return num_pages_with_data;
  }

  Gtk.Widget get_filter_widget ()
  {
    return this.filter;
  }

  public void set_filter_type (FilterType filter_type)
  {
    this.filter_type = filter_type;
  }

  public void set_filter_text (string text)
  {
    if (this.filter_type == FilterType.NONE)
      warning (_("You cannot set the filter text because there is no filter active"));
    else
      this.filter.set_text (text);
  }

  public string get_filter_text ()
  {
    return this.filter.get_text ();
  }

  public void set_key (KeysType type, string keys)
  {
    uint key;
    Gdk.ModifierType mods;
    Gtk.accelerator_parse (keys, out key, out mods);
    return_if_fail (key != 0);
    this.keys[(int)type].key = key;
    this.keys[(int)type].mods = mods;
  }

  public string get_key (KeysType type)
  {
    return Gtk.accelerator_name (this.keys[(int)type].key, this.keys[(int)type].mods);
  }

  public FilterType get_filter_type ()
  {
    return this.filter_type;
  }

  public void bottom_bar_set_visible (bool visible)
  {
    if (visible)
      this.bottom_bar.show ();
    else
      this.bottom_bar.hide ();
  }

  public bool bottom_bar_get_visible ()
  {
    return this.bottom_bar.visible;
  }
}

