/* tree.vala
 *
 * Copyright (C) 2008  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <gege2061@redaction-developpez.com>
 */

/*
 * This code is re-writing in the vala language of GtkSourceCompletion
 *
 * Copyright (C) 2007 - Chuchiperriman
 *
 * rev 163
 */

public class SourceCompletion.Tree : Gtk.ScrolledWindow
{
  private enum Col
  {
    PIXBUF,
    NAME,
    DATA,
    NB
  }

  public signal void proposal_selected (Proposal proposal);
  public signal void selection_changed (Proposal proposal);

  private bool destroy_has_run;
  private string current_filter;
  private bool active_filter;
  private Gtk.TreeView tree_view;
  private Gtk.ListStore list_store;
  private Gtk.TreeModelFilter model_filter;

  private void tree_row_activated_cb (Gtk.TreeView tree_view,
                                      Gtk.TreePath path,
                                      Gtk.TreeViewColumn column)
  {
    Gtk.TreeIter iter;
    Gtk.TreeModel model;
    Proposal proposal;

    model = tree_view.get_model ();

    model.get_iter (out iter, path);
    model.get (iter, Col.DATA, out proposal);

    this.proposal_selected (proposal);
  }

  private void selection_changed_cd (Gtk.TreeSelection treeselection)
  {
    Proposal proposal;

    if (this.get_selected_proposal (out proposal))
    {
      this.selection_changed (proposal);
    }
  }

  private bool filter_by_name_func (Gtk.TreeModel model, Gtk.TreeIter iter)
  {
    if (!this.active_filter)
      return true;
		
    string filter = this.current_filter;
    if (filter == null)
      return true;

    string current_name;
    model.get (iter, Col.NAME, out current_name);

    if (current_name == null)
      return true;

    long len_cur = filter.len ();
    if (filter == current_name.ndup (len_cur))
    {
      return true;
    }
    return false;
  }

  construct
  {
    this.destroy_has_run = false;
    this.current_filter = null;
    this.active_filter = false;
    this.set_policy (Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC);
/*    this.set_hadjustment (null);
    this.set_vadjustment (null);*/

    this.tree_view = new Gtk.TreeView ();
    this.tree_view.can_focus = false;
    this.tree_view.set_headers_visible (false);

    this.add (this.tree_view);

    // Create the Tree
    Gtk.CellRendererText renderer;
    Gtk.TreeViewColumn column;
    Gtk.CellRendererPixbuf renderer_pixbuf;
    Gtk.TreeViewColumn column_pixbuf;

    renderer_pixbuf = new Gtk.CellRendererPixbuf ();
    column_pixbuf = new Gtk.TreeViewColumn.with_attributes ("Pixbuf",
                                                            renderer_pixbuf,
                                                            "pixbuf",
                                                            Col.PIXBUF);
	
    column = new Gtk.TreeViewColumn ();
    renderer = new Gtk.CellRendererText ();
    column.pack_start (renderer, false);
    column.set_attributes (renderer, "text", Col.NAME);

    this.tree_view.append_column (column_pixbuf);
    this.tree_view.append_column (column);

    // Create the model
    this.list_store = new Gtk.ListStore (Col.NB, typeof (Gdk.Pixbuf),
                                                 typeof (string),
                                                 typeof (Proposal));
	
    this.model_filter = new Gtk.TreeModelFilter (this.list_store, null);

    this.model_filter.set_visible_func (filter_by_name_func);

    this.tree_view.set_model (this.model_filter);
	
	
    // Connect signals
    this.tree_view.row_activated += tree_row_activated_cb;

    Gtk.TreeSelection selection = this.tree_view.get_selection ();
    selection.changed += selection_changed_cd;
  }

  public bool get_selected_proposal (out Proposal proposal)
  {
    Gtk.TreeIter iter;
    Gtk.TreeModel model;
    Gtk.TreeSelection selection = this.tree_view.get_selection ();

    if (selection.get_selected (null, out iter))
    {
      model = this.tree_view.get_model ();
      model.get (iter, Col.DATA, out proposal);
      return true;
    }
    return false;
  }

  public void add_data (Proposal data)
  {
    Gtk.TreeIter iter;
    Gtk.ListStore store = this.list_store;

    store.append (out iter);

    store.set (iter, Col.PIXBUF, data.icon,
                     Col.NAME, data.label,
                     Col.DATA, data);
  }

  public void clear ()
  {
    this.list_store.clear ();
  }

  public bool select_first ()
  {
    Gtk.TreeIter iter;
    Gtk.TreePath path;
    Gtk.TreeModel model;
    Gtk.TreeSelection selection;

    if (!this.visible)
      return false;
	
    selection = this.tree_view.get_selection ();

    if (selection.get_mode () == Gtk.SelectionMode.NONE)
      return false;

    model = this.tree_view.get_model ();

    if (model.get_iter_first (out iter))
    {
      selection.select_iter (iter);
      path = model.get_path (iter);
      this.tree_view.scroll_to_cell (path, null, false, 0, 0);
      return true;
    }
    else
    {
      return false;
    }
  }

  public bool select_last ()
  {
    Gtk.TreeIter iter;
    Gtk.TreeModel model;
    Gtk.TreeSelection selection;
    Gtk.TreePath path;
    int children;

    if (!this.visible)
      return false;

    selection = this.tree_view.get_selection ();
    model = this.tree_view.get_model ();

    if (selection.get_mode () == Gtk.SelectionMode.NONE)
      return false;

    children = model.iter_n_children (null);
    if (children > 0)
    {
      model.iter_nth_child (out iter, null, children);

      selection.select_iter (iter);
      path = model.get_path (iter);
      this.tree_view.scroll_to_cell (path, null, false, 0, 0);
      return true;
    }
    return false;
  }

  public bool select_previous (int rows)
  {
    Gtk.TreeIter iter;
    Gtk.TreePath path;
    Gtk.TreeModel model;
    Gtk.TreeSelection selection;

    if (!this.visible)
      return false;

    selection = this.tree_view.get_selection ();

    if (selection.get_mode () == Gtk.SelectionMode.NONE)
      return false;

    if (selection.get_selected (out model, out iter))
    {
      path = model.get_path (iter);
      for (int i=0; i < rows; i++)
      {
        path.prev ();
      }

      if (model.get_iter(out iter, path))
      {
        selection.select_iter (iter);
        this.tree_view.scroll_to_cell (path, null, false, 0, 0);
      }
    }
    else
    {
      return this.select_first();
    }
    return true;
  }

  public bool select_next (int rows)
  {
    Gtk.TreeIter iter;
    Gtk.TreeModel model;
    Gtk.TreeSelection selection;

    if (!this.visible)
      return false;

    selection = this.tree_view.get_selection ();

    if (selection.get_mode () == Gtk.SelectionMode.NONE)
      return false;

    if (selection.get_selected (out model, out iter))
    {
      Gtk.TreePath path = model.get_path (iter);
      for (int i=0; i < rows; i++)
      {
        if (!model.iter_next (ref iter))
          return this.select_last ();
      }

      selection.select_iter (iter);
      path = model.get_path (iter);
      this.tree_view.scroll_to_cell (path, null, false, 0, 0);
    }
    else
    {
      return this.select_first();
    }
    return true;
  }

  public int get_num_proposals ()
  {
    Gtk.TreeModel model = this.tree_view.get_model ();
    return model.iter_n_children (null);
  }

  public void filter (string filter)
  {
    this.active_filter = true;
    this.current_filter = filter;
    this.model_filter.refilter ();
    this.active_filter = false;
    this.current_filter = null;
    //this.select_first ();
  }
}

