/* opened-documents.vala
 *
 * Copyright (C) 2009 Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <gege2061@redaction-developpez.com>
 */

using Valide;

public class OpenedDocuments : Plugin
{
  /* enum for tree model columns, not all is currently used */
  enum Col
  {
    ICON,
    NAME,
    DOCUMENT,
    NB
  }

  private Gtk.VBox box;
  private Gtk.TreeView tree_view;
  private DocumentManager documents;

  private bool populate_asc ()
  {
    Gtk.TreeIter iter;
    Gtk.ListStore model = (Gtk.ListStore)this.tree_view.get_model ();
    model.clear ();

    foreach (iDocument document in this.documents.documents)
    {
      model.append (out iter);
      model.set (iter, Col.ICON, document.get_icon (),
                 Col.NAME, document.filename,
                 Col.DOCUMENT, document);
    }
    this.on_active_tab_changed (this.documents, null, 0, this);
    return false;
  }

  [Callback]
  private void populate ()
  {
    Idle.add (this.populate_asc);
  }

  [Callback]
  private void on_cursor_change (Gtk.TreeView sender)
  {
    iDocument document;
    Gtk.TreeModel model;
    Gtk.TreeIter iter;

    Gtk.TreeSelection selection = this.tree_view.get_selection ();

    if (selection.get_selected (out model, out iter))
    {
      model.get (iter, Col.DOCUMENT, out document);
      DocumentManager.get_instance ().current = document;
    }
  }

  [Callback]
  private static void on_active_tab_changed (Gtk.Notebook sender,
                                             Gtk.NotebookPage? page,
                                             uint page_num,
                                             OpenedDocuments self)
  {
    Gtk.TreeIter iter;
    iDocument document;
    Gtk.ListStore list_store = (Gtk.ListStore)self.tree_view.get_model ();

    if (list_store.get_iter_first (out iter))
    {
      do
      {
        list_store.get (iter, Col.DOCUMENT, out document);
        if (document == self.documents.current)
        {
          self.tree_view.get_selection ().select_iter (iter);
          break;
        }
      } while (list_store.iter_next (ref iter));
    }
  }

  construct
  {
    this.box = new Gtk.VBox (false, 0);

    /* create some nice padding */
    this.box.set_spacing (2);
    this.box.set_border_width (2);
 
    /* create and pack tree view */
    Gtk.ScrolledWindow scrolled = new Gtk.ScrolledWindow (null, null);
    scrolled.set_policy (Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC);
    scrolled.set_shadow_type (Gtk.ShadowType.IN);
    this.box.pack_start (scrolled, true, true, 0);

    this.tree_view = new Gtk.TreeView ();
    this.tree_view.set_rules_hint (false);
    this.tree_view.set_headers_visible (false);
    scrolled.add (this.tree_view);

    /* create model */
    Gtk.ListStore store = new Gtk.ListStore (Col.NB,
                                             typeof (Gdk.Pixbuf),
                                             typeof (string),
                                             typeof (iDocument));
    store.set_sort_column_id (Col.NAME, Gtk.SortType.ASCENDING);

    /* main column */
    Gtk.TreeViewColumn col = new Gtk.TreeViewColumn ();
    col.set_title(_("File"));
    {
      Gtk.CellRendererPixbuf renderer = new Gtk.CellRendererPixbuf ();
      col.pack_start (renderer, false);
      col.set_attributes (renderer, "pixbuf", Col.ICON);
    }
    {
      Gtk.CellRendererText renderer = new Gtk.CellRendererText ();
      col.pack_start (renderer, false);
      col.set_attributes (renderer, "markup", Col.NAME);
    }
    this.tree_view.append_column (col);
    this.tree_view.set_model (store);

    /* attach signal callbacks for tree view */
    this.tree_view.cursor_changed += this.on_cursor_change;

    this.documents = DocumentManager.get_instance ();

    this.documents.tab_removed += this.populate;
    this.documents.tab_added += this.populate;
    this.documents.tab_state_changed += populate;
    Signal.connect_after (this.documents, "switch-page", (Callback)on_active_tab_changed, this);

    Window.get_instance ().side_panel.add_item_with_stock_icon (this.box,
                                                                _("Documents"),
                                                                Gtk.STOCK_FILE);

    this.populate ();
    this.box.show_all ();
  }

  ~OpenedDocuments ()
  {
    this.documents.tab_removed -= this.populate;
    this.documents.tab_added -= this.populate;
    this.documents.tab_state_changed -= populate;
    SignalHandler.disconnect_by_func (this.documents, (void*)on_active_tab_changed, this);
    Window.get_instance ().side_panel.remove_item (this.box);
  }
}

[ModuleInit]
public Type register_plugin ()
{
  return typeof (OpenedDocuments);
}

