/* symbol-browser.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <gege2061@redaction-developpez.com>
 */

/*
 * This code is re-writing in the vala language of GeditSymbolBrowserPlugin
 *
 * Copyright (C) 2007 - Micah Carrick
 */

using Valide;

/* just used to keep the population code readable */
private class Symbol
{
  public string name;
  public string type;
  public string file;
  public string language;
  public int line;
  public Gtk.TreeIter iter;
  public Symbol root;
  public weak AbstractTextDocument document;
}

public class SymbolBrowser : Plugin
{
  /* enum for tree model columns, not all is currently used */
  enum Col
  {
    ICON,
    NAME,
    FILE,
    LINE,
    LANGUAGE,
    DOCUMENT,
    NB
  }

  private Gtk.VBox box;
  private Gtk.TreeView tree_view;
  private Gtk.Label label;
  private DocumentManager documents;
  private SList<string> expanded_rows;
  private uint status_context;
  private Gtk.TreeViewColumn language_column;
  private Gtk.TreeViewColumn source_file_column;
  private Gtk.TreeViewColumn line_number_column;
  private HashTable<string, Gdk.Pixbuf>icons;

  private bool show_language_column
  {
    get
    {
      return ConfigManager.get_instance ().get_boolean ("SymbolBrowser", "show-language");
    }
    set
    {
      ConfigManager.get_instance ().set_boolean ("SymbolBrowser", "show-language",
                                                 value);
      this.language_column.visible = value;
    }
  }

  private bool show_source_file_column
  {
    get
    {
      return ConfigManager.get_instance ().get_boolean ("SymbolBrowser", "show-source-file");
    }
    set
    {
      ConfigManager.get_instance ().set_boolean ("SymbolBrowser", "show-source-file",
                                                 value);
      this.source_file_column.visible = value;
    }
  }

  private bool show_line_number_column
  {
    get
    {
      return ConfigManager.get_instance ().get_boolean ("SymbolBrowser", "show-line-number");
    }
    set
    {
      ConfigManager.get_instance ().set_boolean ("SymbolBrowser", "show-line-number",
                                                 value);
      this.line_number_column.visible = value;
    }
  }

  private bool only_load_active_tab
  {
    get
    {
      return true;
    }
  }

  private bool hierarchic_view
  {
    get
    {
      return ConfigManager.get_instance ().get_boolean ("SymbolBrowser", "hierarchic-view");
    }
    set
    {
      ConfigManager.get_instance ().set_boolean ("SymbolBrowser", "hierarchic-view", value);
      this.populate_symbols (this, this);
    }
  }

  private void save_expanded_rows ()
  {
/*
        GConfClient             *client;
        GError                  *error=NULL;
        
        client = gconf_client_get_default ();
        
        gconf_client_set_list (client, 
                        GEDIT_SYMBOL_BROWSER_PLUGIN_GCONF_PATH"expanded-rows",
                        GCONF_VALUE_STRING, panel->priv->expanded_rows,
                        &error);
                        
        if (error != NULL)
        {
                g_warning ("Error saving gedit-symbol-browser-plugin setting: %s", 
                                error->message);
                                
                g_error_free (error);
        }

        g_object_unref (client);
*/
  }

  private bool is_expanded_row (string name)
  {
    bool ret = false;

    foreach (string i in this.expanded_rows)
    {
      if (i == name)
      {
        ret = true;
        break;
      }
    }
    return ret;
  }

  private void add_expanded_row (string name)
  {
    if (!this.is_expanded_row (name))
    {
      this.expanded_rows.prepend (name);
      this.save_expanded_rows ();
    }
  }

  private void remove_expanded_row (string name)
  {
    weak SList<string> l;

    if (this.is_expanded_row (name) == true)
    {
      for (l = this.expanded_rows; l != null; l = l.next)
      {
        if (l.data == name)
        {
          this.expanded_rows.remove_all (l.data);
        }
      }
      this.save_expanded_rows ();
    }
  }

  /**
   * Execute the ctags command and return stdout.
   */ 
  private string exec_ctags (string filename) throws Error
  {
    string output;

    /* build ctags command */
    string command = "%s  -n --sort=no --fields=-k-f-s-t+K+l+n -f - \"%s\""
                       .printf (Config.VALIDE_CTAGS_EXEC, filename);
    /* execute command */
    Process.spawn_command_line_sync (command, out output, null, null);

    return output;
  }

  /**
   * This function needs work. Used to be custom parsed for ex searches
   * which could potentially contain tabs, however, I've decided that isn't
   * really needed. But, it's still using Symbol. We could just pass back
   * **fields array.
   */ 
  private Symbol parse_line (string line)
  {
    /* name, file, ex search, type, line:n  language:x */

    Symbol symbol = new Symbol ();

    string[] fields = line.split ("\t", 0);
    if (fields[0] != null)
    {
      symbol.name = fields[0];

      symbol.file = Path.get_basename (fields[1]);
      symbol.type = fields[3];

      string[] splits = fields[4].split(":", 0);
      symbol.line = splits[1].to_int ();

      splits = fields[5].split (":", 0);
      symbol.language = splits[1];
      symbol.root = null;
    }
    return symbol;
  }

  /**
   * Return the PixBuf that represents the specified type of symbol, or NULL.
   */ 
  private Gdk.Pixbuf get_symbol_pixbuf (string type)
  {
    Gdk.Pixbuf pixbuf = this.icons.lookup (type);

    if (pixbuf == null)
    {
      try
      {
        string filename = "%s.png".printf (type);
        string path = Path.build_filename (this.path, "pixmaps", filename);

        pixbuf = new Gdk.Pixbuf.from_file (path);
        this.icons.insert (type, pixbuf);
      }
      catch (Error e)
      {
        debug ("Could not load pixbuf: %s\n", e.message);
        pixbuf = Utils.get_pixbuf_for_theme (Gtk.STOCK_MISSING_IMAGE, Gtk.IconSize.MENU);
      }
    }
    return pixbuf;
  }

  /**
   * Add a single symbol to the treeview. Parents rows are added as the type
   * whenever they aren't already present.
   */
  private void add_symbol (Symbol symbol)
  {
    Gtk.TreeIter iter;
    Gtk.TreeView tree_view = this.tree_view;
    Gdk.Pixbuf pixbuf = this.get_symbol_pixbuf (symbol.type);
    Gtk.TreeStore store = (Gtk.TreeStore)tree_view.get_model ();

    if (!this.hierarchic_view)
    {
      /* check for existing parent iter */
      Gtk.TreeIter parent = Gtk.TreeIter ();
      bool valid = true;
      bool no_parent = true;

      /* we could do bold here */
      string markup = "<b>%s</b>".printf (symbol.type);

      for (int i = 0; (valid && no_parent); i++)
      {
        string path_string = i.to_string ();
        valid = store.get_iter_from_string (out parent, path_string);
        if (valid)
        {
          string val;
          store.get (parent, Col.NAME, out val, -1);
          if (val == markup) 
          {
            no_parent = false;
          }
        }
      }

      /* create the parent if it wasn't already found */
      if (no_parent)
      {
        store.append (out parent, null);
        store.set (parent, Col.ICON, pixbuf, Col.NAME, markup, -1);
      }
      store.append (out iter, parent);
    }
    else
    {
      if (symbol.root != null)
      {
        store.append (out iter, symbol.root.iter);
      }
      else
      {
        store.append (out iter, null);
      }
    }

    /* add the new symbol */
    store.set (iter, Col.ICON, pixbuf,
                     Col.NAME, symbol.name,
                     Col.FILE, symbol.file,
                     Col.LINE, symbol.line,
                     Col.LANGUAGE, symbol.language,
                     Col.DOCUMENT, symbol.document,
                     -1);
    symbol.iter = iter;
  }

  private bool populate_symbols_asc ()
  {
    this.box.sensitive = false;
    /* clear data */
    Gtk.TreeStore model = (Gtk.TreeStore)this.tree_view.get_model ();
    model.clear ();
    this.label.set_text ("");

    if (this.documents.current != null)
    {
      int count = 0;

      /* find out if we want to load all symbols or just active tab */
      List<iDocument> doc = null;
      if (this.only_load_active_tab)
      {
        doc.prepend (this.documents.current);
      }
      else
      {
        doc = this.documents.documents.copy ();
      }

      /* run ctags on each document in the list and add symbols */
      foreach (iDocument d in doc)
      {
        if (d is AbstractTextDocument && d.path != null
            && FileUtils.test (d.path, FileTest.EXISTS))
        {
          try
          {
            string output = this.exec_ctags (d.path);

            Symbol root = null;
            string[] lines = output.split ("\n", 0);
            for (int i = 0; lines[i] != null; i++)
            {
              if (lines[i] != "")
              {
                Symbol symbol = this.parse_line (lines[i]);
                if (symbol != null)
                {
                  symbol.document = (AbstractTextDocument)d;
                  if (symbol.type == "class")
                  {
                    root = null;
                  }
                  else if (root != null
                           && ((root.type == "enum" && symbol.type != "enumvalue")
                           || (root.type == "errordomain" && symbol.type != "errorcode")))
                  {
                    root = root.root;
                  }
                  symbol.root = root;
                  this.add_symbol (symbol);
                  if (symbol.type == "class" || symbol.type == "enum"
                      || symbol.type == "errordomain")
                  {
                    root = symbol;
                  }
                }
              }
            }
            count++;
          }
          catch (Error e)
          {
            debug (_("Could not execute ctags: %s\n"), e.message);
          }
        }
      }
//       this.restore_expanded_rows ();
      this.tree_view.expand_all ();
      if (count > 0)
      {
        this.box.sensitive = true;
      }
    }
    return false;
  }

  /**
   * Call to re-populate tree view.
   */
  [Callback]
  private static void populate_symbols (Object sender, SymbolBrowser self)
  {
    Idle.add (self.populate_symbols_asc);
  }

  /**
   * Re-expand any row when the list is repopulated if that row was previously
   * expanded.
   */
  private void restore_expanded_rows ()
  {
    /* is this going to fix the crash? */
    if (this.documents.sensitive == true)
    {
      Gtk.TreeIter iter;
      Gtk.TreeModel model = this.tree_view.get_model ();

      if (model.get_iter_first (out iter) == true)
      {
        for (bool r = model.get_iter_first (out iter); 
             r == true;
             r = model.iter_next (ref iter))
        {
          string name;

          model.get (iter, Col.NAME, out name,  -1);
          if (this.is_expanded_row (name))
          {
            Gtk.TreePath path = model.get_path (iter);
            this.tree_view.expand_row (path, false);
          }
        }
      }
    }
  }

  [Callback]
  private static void on_active_tab_changed (DocumentManager sender, Gtk.NotebookPage page, uint page_num, SymbolBrowser self)
  {
    /* only needed in single-file mode  */
    if (self.only_load_active_tab)
    {
      populate_symbols (sender, self);
    }
  }

  [Callback]
  private void on_row_activated (Gtk.TreeView sender, Gtk.TreePath path,
                                 Gtk.TreeViewColumn column)
  {
    Gtk.TreeIter iter;
        
    Gtk.TreeModel model = this.tree_view.get_model ();
    model.get_iter (out iter, path);

    /* expand/collapse row if it has children (and return) */
    if (!this.hierarchic_view && model.iter_has_child (iter))
    {
      if (this.tree_view.row_is_expanded (path))
      {
        this.tree_view.collapse_row (path);
      }
      else
      {
        this.tree_view.expand_row (path, false);
      }
    }
    else
    {
      /* otherwise, show the file name and line number and flash statusbar */
      AbstractTextDocument document;
      string name;
      int line;

      model.get (iter, Col.NAME, out name,  Col.LINE, out line, Col.DOCUMENT,
                 out document, -1);

      DocumentManager.get_instance ().current = document;
      document.goto_line (line - 1);
      document.grab_focus ();
    }
  }

  [Callback]
  private void on_cursor_changed (Gtk.TreeView sender)
  {
    Gtk.TreeIter iter;
    Gtk.TreeModel model = this.tree_view.get_model ();
    Gtk.TreeSelection selection = this.tree_view.get_selection ();

    if (selection.get_selected (null, out iter))
    {
      AbstractTextDocument document;
      string name;
      int line;
      string file;

      model.get (iter, Col.NAME, out name, Col.FILE, out file,
                       Col.DOCUMENT, out document, Col.LINE, out line, -1);

			if (file != null)
			{
		    string label_text = "%s Line:%d".printf (file, line);
		    this.label.set_text (label_text);

		    string status_message = "Double-click to jump to '%s'".printf (name);

		    Statusbar.get_instance ().flash_message (this.status_context, status_message);
		  }
    }
  }

  [Callback]
  private void on_row_expanded (Gtk.TreeView sender, Gtk.TreeIter iter,
                                Gtk.TreePath path)
  {
    string name;        

    Gtk.TreeModel model = this.tree_view.get_model ();
    model.get (iter, Col.NAME, out name, -1);
    return_if_fail (name != null);

    this.add_expanded_row (name);
  }

  [Callback]
  private void on_row_collapsed (Gtk.TreeView sender, Gtk.TreeIter iter,
                                 Gtk.TreePath path)
  {
    string name;

    Gtk.TreeModel model = this.tree_view.get_model ();
    model.get (iter, Col.NAME, out name, -1);
    return_if_fail (name != null);

    remove_expanded_row (name);
  }

  construct
  {
    this.box = new Gtk.VBox (false, 0);

    /* create some nice padding */
    this.box.set_spacing (2);
    this.box.set_border_width (2);
 
    this.icons = new HashTable<string, Gdk.Pixbuf> (str_hash, str_equal);
 
    /* create and pack tree view */
    Gtk.ScrolledWindow scrolled = new Gtk.ScrolledWindow (null, null);
    scrolled.set_policy (Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC);
    scrolled.set_shadow_type (Gtk.ShadowType.IN);
    this.box.pack_start (scrolled, true, true, 0);

    this.tree_view = new Gtk.TreeView ();
    this.tree_view.set_rules_hint (false);
    this.tree_view.set_headers_visible (true);
    this.tree_view.set_headers_clickable (true);
    scrolled.add (this.tree_view);

    /* create and pack label */
    this.label = new Gtk.Label ("");
    this.label.set_alignment ((float)0.0, (float)0.5);
    this.box.pack_start (this.label, false, true, 0);

    /* create model */
    Gtk.TreeStore store = new Gtk.TreeStore (Col.NB,
                                             typeof (Gdk.Pixbuf),
                                             typeof (string),
                                             typeof (string),
                                             typeof (int),
                                             typeof (string),
                                             typeof (Object));
    store.set_sort_column_id (1, Gtk.SortType.ASCENDING);

    /* main column */
    Gtk.TreeViewColumn col = new Gtk.TreeViewColumn ();
    col.set_title(_("Symbol"));
    {
      Gtk.CellRendererPixbuf renderer = new Gtk.CellRendererPixbuf ();
      col.pack_start (renderer, false);
      col.set_attributes (renderer, "pixbuf", Col.ICON);
    }
    {
      Gtk.CellRendererText renderer = new Gtk.CellRendererText ();
      col.pack_start (renderer, false);
      col.set_attributes (renderer, "markup", Col.NAME);
    }
    this.tree_view.append_column (col);

    /* additional columns */

    this.language_column = new Gtk.TreeViewColumn ();
    this.show_language_column = ConfigManager.get_instance ().get_boolean ("SymbolBrowser", "show-language");
    this.language_column.set_title (_("Language"));
    Gtk.CellRendererText renderer = new Gtk.CellRendererText ();
    this.language_column.pack_start (renderer, true);
    this.language_column.set_attributes (renderer, "text", Col.LANGUAGE);
    this.tree_view.append_column (this.language_column);

    this.source_file_column = new Gtk.TreeViewColumn ();
    this.show_source_file_column = ConfigManager.get_instance ().get_boolean ("SymbolBrowser", "show-source-file");
    this.source_file_column.set_title (_("File"));
    renderer = new Gtk.CellRendererText ();
    this.source_file_column.pack_start (renderer, true);
    this.source_file_column.set_attributes (renderer, "text", Col.FILE);
    this.tree_view.append_column (this.source_file_column);

    this.line_number_column = new Gtk.TreeViewColumn ();
    this.show_line_number_column = ConfigManager.get_instance ().get_boolean ("SymbolBrowser", "show-line-number");
    this.line_number_column.set_title (_("Line"));
    renderer = new Gtk.CellRendererText ();
    this.line_number_column.pack_start (renderer, true);
    this.line_number_column.set_attributes (renderer, "text", Col.LINE);
    this.tree_view.append_column (this.line_number_column);

    this.tree_view.set_model (store);

    /* attach signal callbacks for tree view */
    this.tree_view.row_activated += on_row_activated;
    this.tree_view.cursor_changed += on_cursor_changed;
    this.tree_view.row_expanded += on_row_expanded;
    this.tree_view.row_collapsed += on_row_collapsed;

    this.status_context = Statusbar.get_instance ().get_context_id ("valide_symbol_browser");

    this.documents = DocumentManager.get_instance ();
    this.box.sensitive = false;

    Signal.connect_after (this.documents, "tab-removed", (Callback)populate_symbols, this);
    Signal.connect_after (this.documents, "switch-page", (Callback)on_active_tab_changed, this);
    Signal.connect_after (this.documents, "tab-saved", (Callback)populate_symbols, this);

    Gtk.Image image = new Gtk.Image.from_file (Path.build_filename (this.path,
                                                                    "symbol-browser.png"));

    Window.get_instance ().side_panel.add_item (this.box, "Symbol Browser", image);

    this.box.show_all ();
  }

  public SymbolBrowser (string path)
  {
    this.path = path;
  }

  ~SymbolBrowser ()
  {
    SignalHandler.disconnect_by_func (this.documents, (void*)populate_symbols, this);
    SignalHandler.disconnect_by_func (this.documents, (void*)on_active_tab_changed, this);
    Window.get_instance ().side_panel.remove_item (this.box);
  }

	public Gtk.Widget create_configure_dialog ()
	{
	  Gtk.Frame frame = new Gtk.Frame (_("Additionnal columns"));

	  Gtk.VBox vbox = new Gtk.VBox (true, 0);
	  frame.add (vbox);

	  Gtk.CheckButton check = new Gtk.CheckButton.with_label (_("Programming language"));
	  check.set_active (this.show_language_column);
	  check.toggled += (s) => {
	    this.show_language_column = s.get_active ();
	  };
	  vbox.pack_start (check, true, true, 0);

	  check = new Gtk.CheckButton.with_label (_("Source file"));
	  check.set_active (this.show_source_file_column);
	  check.toggled += (s) => {
	    this.show_source_file_column = s.get_active ();
	  };
	  vbox.pack_start (check, true, true, 0);

	  check = new Gtk.CheckButton.with_label (_("Line number"));
	  check.set_active (this.show_line_number_column);
	  check.toggled += (s) => {
	    this.show_line_number_column = s.get_active ();
	  };
	  vbox.pack_start (check, true, true, 0);

	  check = new Gtk.CheckButton.with_label (_("Hierarchic view"));
	  check.set_active (this.hierarchic_view);
	  check.toggled += (s) => {
	    this.hierarchic_view = s.get_active ();
	  };
	  vbox.pack_start (check, true, true, 0);

	  return frame;
	}
}

[ModuleInit]
public Type register_plugin ()
{
  return typeof (SymbolBrowser);
}

public Gtk.Widget create_configure_dialog (SymbolBrowser self)
{
  return self.create_configure_dialog ();
}

